/* -*- Mode: C++ -*-
 * $Id: Sensors.cc,v 1.22 2001/11/15 16:48:48 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.cc
 * Unitname: BasicSystem
 * $Revision: 1.22 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2001/11/15 16:48:48 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.cc,v $
 * Revision 1.22  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.21  2001/10/14 21:21:43  frehe
 * Added better support for messages from self and from other players.
 *
 * Revision 1.20  2001/10/14 14:48:36  frehe
 * Removed old debug info.
 *
 * Revision 1.19  2001/09/22 17:18:28  frehe
 * Added a workaround to get teammates from rccparser.
 *
 * Revision 1.18  2001/09/17 21:53:59  frehe
 * Readded support for a debug coach.
 *
 * Revision 1.17  2001/09/17 21:30:09  andbo928
 * Implemented an ugly workaround to get ballspeeds working.
 *
 * Revision 1.16  2001/09/14 13:48:04  frehe
 * Fixed the speed bug.
 *
 * Revision 1.15  2001/09/03 19:39:36  frehe
 * Changed code to use the RS_BEGIN_TRY, RS_END_TRY and RS_CATCH macros.
 *
 * Revision 1.14  2001/09/01 23:52:23  frehe
 * Added some debug info.
 *
 * Revision 1.13  2001/08/30 21:41:47  frehe
 * Removed a call to distDirChng in RCCParser that crashed the agent.
 *
 * Revision 1.12  2001/08/29 16:37:32  frehe
 * Added a test on the validity of vec in getVectorUnknown(RCC_Vector2d).
 *
 * Revision 1.11  2001/08/29 12:35:26  frehe
 * Fixed a bug in sense_body handler.
 *
 * Revision 1.10  2001/08/22 21:06:59  frehe
 * Updated 2001-08-22
 *
 * Revision 1.9  2001/08/19 23:32:24  frehe
 * Updated 2001-08-19
 *
 * Revision 1.8  2001/08/17 07:15:16  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.7  2001/08/13 16:43:16  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.6  2001/08/13 07:33:16  frehe
 * Updated from laptop 2001-08-13.
 *
 * Revision 1.5  2001/08/02 17:04:22  frehe
 * Updated 2001-08-02
 *
 * Revision 1.4  2000/10/24 15:15:17  frehe
 * Fixed the generation of the onNewCycle event.
 *
 * Revision 1.3  2000/09/05 21:15:53  frehe
 * Made minor changes.
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:14  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.cc
 * Implementation file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.22 $
 * $Author: frehe $
 * $Date: 2001/11/15 16:48:48 $
 */


#include <cstdio>
#include <cstring>
#include <cctype>

#include "Sensors.h"
#include "Controller.h"
#include "SensorInterface.h"
#include "string_utils.h"
#include "rs_error.h"
#include "rs_log.h"
#include "rs_try_catch.h"


RS_BEGIN_NAMESPACE


SensorDataHandler::SensorDataHandler(Sensors* _sensors)
  : sensors(_sensors)
{
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_CoachAudio>& ca_p)
{
  setLastSensorType(SENSOR_AURAL_COACH);
  updateTime(ca_p->time());
  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Heard from coach!");
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_Error>& error_p)
{
  setLastSensorType(SENSOR_ERROR);
  RS_ERROR("Sent an illegal message");
}


PlayMode
SensorDataHandler::getPlayMode(const RCC_StadPlayMode& play_mode) const
{
  PlayMode pm = PM_UNKNOWN;
  switch (play_mode.mode()) {
  case RCC_StadPlayMode::Unknown:
    pm = PM_UNKNOWN;
    break;
  case RCC_StadPlayMode::BeforeKickOff:
    pm = PM_BEFORE_KICK_OFF;
    break;
  case RCC_StadPlayMode::TimeOver:
    pm = PM_TIME_OVER;
    break;
  case RCC_StadPlayMode::PlayOn:
    pm = PM_PLAY_ON;
    break;
  case RCC_StadPlayMode::KickOff_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_KICK_OFF:PM_THEIR_KICK_OFF;
    break;
  case RCC_StadPlayMode::KickIn_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_KICK_IN:PM_THEIR_KICK_IN;
    break;
  case RCC_StadPlayMode::FreeKick_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FREE_KICK:PM_THEIR_FREE_KICK;
    break;
  case RCC_StadPlayMode::CornerKick_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_CORNER_KICK:PM_THEIR_CORNER_KICK;
    break;
  case RCC_StadPlayMode::GoalKick_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_GOAL_KICK:PM_THEIR_GOAL_KICK;
    break;
  case RCC_StadPlayMode::AfterGoal_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_GOAL:PM_THEIR_GOAL;
    break;
  case RCC_StadPlayMode::Drop_Ball:
    pm = PM_DROP_BALL;
    break;
  case RCC_StadPlayMode::OffSide_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_OFFSIDE_KICK:PM_THEIR_OFFSIDE_KICK;
    break;
  case RCC_StadPlayMode::PK_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_PENALTY_KICK:PM_THEIR_PENALTY_KICK;
    break;
  case RCC_StadPlayMode::FirstHalfOver:
    pm = PM_FIRST_HALF_OVER;
    break;
  case RCC_StadPlayMode::Pause:
    pm = PM_PAUSE;
    break;
  case RCC_StadPlayMode::Human:
    pm = PM_HUMAN;
    break;
  case RCC_StadPlayMode::Foul_Charge_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FOUL:PM_THEIR_FOUL;
    break;
  case RCC_StadPlayMode::Foul_Push_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FOUL:PM_THEIR_FOUL;
    break;
  case RCC_StadPlayMode::Foul_MultipleAttacker_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FOUL:PM_THEIR_FOUL;
    break;
  case RCC_StadPlayMode::Foul_BallOut_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FOUL:PM_THEIR_FOUL;
    break;
  case RCC_StadPlayMode::GoalieCatch_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_GOALIE_GOT_BALL:PM_THEIR_GOALIE_GOT_BALL;
    break;
  case RCC_StadPlayMode::TimeUpWithoutTeam:
    pm = PM_TIME_UP_WITHOUT_TEAM;
    break;
  case RCC_StadPlayMode::TimeUp:
    pm = PM_TIME_UP;
    break;
  case RCC_StadPlayMode::TimeExtended:
    pm = PM_EXTENDED_TIME;
    break;
  case RCC_StadPlayMode::HalfTime:
    pm = PM_HALF_TIME;
    break;
  case RCC_StadPlayMode::Foul_:
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_FOUL:PM_THEIR_FOUL;
    break;
  case RCC_StadPlayMode::Goal_:
    // Not used I think...
    pm = (play_mode.side().mode() == RCC_Side::Our)?PM_OUR_GOAL:PM_THEIR_GOAL;
    break;
  default:
    pm = PM_UNKNOWN;
  }

  return pm;
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_InitMsg>& init_p)
{
  setLastSensorType(SENSOR_INIT);

  char side;
  switch ( init_p->side() ) {
  case LEFT:
    side = 'l';
    break;
  case RIGHT:
    side = 'r';
    break;
  default:
    side = 'u';
    RS_ERROR("Unknown side in init message!");
  }
  
  sensors->theSensorInterface->hearInitMessage(side, init_p->uNum(),
					       getPlayMode(init_p->playMode()),
					       init_p->reconnect());
  
  printf( "------------------------------------------------------------------------\n");
  printf( "Connected player %d on team %s\n", init_p->uNum(),
	  sensors->theSensorInterface->getOurTeamName().c_str());
  printf( "He's playing on the %c side of the field.\n", side);
  printf( "------------------------------------------------------------------------\n");

  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Got Init Message!");
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_PlayerAudio>& pa_p)
{
  setLastSensorType(SENSOR_AURAL_PLAYER);
  updateTime(pa_p->time());
  std::string str = pa_p->msg();
  sensors->theSensorInterface->hearPlayer(pa_p->dir().deg(), str.c_str());
  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Heard from player!");
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_PlayerParam>& pp_p)
{
  setLastSensorType(SENSOR_PLAYER_PARAM);
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_PlayerType>& pt_p)
{
  setLastSensorType(SENSOR_PLAYER_TYPE);
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_RefAudio>& ra_p)
{
  setLastSensorType(SENSOR_AURAL_REFEREE);
  updateTime(ra_p->time());
  PlayMode pm = getPlayMode(ra_p->playMode());
  sensors->theSensorInterface->hearReferee(pm);
  if ( pm == PM_OUR_GOAL ) {
    sensors->theSensorInterface->hearScore(sensors->theSensorInterface->getOurSide(),
					   ra_p->playMode().value());
  } else if ( pm == PM_THEIR_GOAL ) {
    sensors->theSensorInterface->hearScore(sensors->theSensorInterface->getTheirSide(),
					   ra_p->playMode().value());
  }
  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Heard from referee!");
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_Score>& score_p)
{
  setLastSensorType(SENSOR_AURAL_SCORE);
  sensors->theSensorInterface->hearScore(sensors->theSensorInterface->getOurSide(),
					 score_p->ourScore());
  sensors->theSensorInterface->hearScore(sensors->theSensorInterface->getTheirSide(),
					 score_p->oppScore());
  updateTime(score_p->time());
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_SelfAudio>& sa_p)
{
  setLastSensorType(SENSOR_AURAL_SELF);
  updateTime(sa_p->time());
  std::string str = sa_p->msg();
  sensors->theSensorInterface->hearSelf(str.c_str());
  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Heard from self!");
}


ViewQuality
SensorDataHandler::getViewQuality(const RCC_ViewMode::quality_t& rcc_view_quality) const
{
  // Conversion between RCC_Parser and RoboSoc types
  ViewQuality view_quality = VQ_UNKNOWN;
  switch ( rcc_view_quality ) {
  case RCC_ViewMode::HIGH:
    view_quality = VQ_HIGH;
    break;
  case RCC_ViewMode::LOW:
    view_quality = VQ_LOW;
    break;
  default:
    RS_ERROR("Unknown view quality");
  }
  return view_quality;
}


ViewWidth
SensorDataHandler::getViewWidth(const RCC_ViewMode::width_t& rcc_view_width) const
{
  // Conversion between RCC_Parser and RoboSoc types
  ViewWidth view_width = VW_UNKNOWN;
  switch ( rcc_view_width ){
  case RCC_ViewMode::NORMAL:
    view_width = VW_NORMAL;
    break;
  case RCC_ViewMode::NARROW:
    view_width = VW_NARROW;
    break;
  case RCC_ViewMode::WIDE:
    view_width = VW_WIDE;
    break;
  default:
    RS_ERROR("Unknown view width");
  }
  return view_width;
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_SenseBody>& sb_p)
{
  setLastSensorType(SENSOR_SENSE);

  static UInt16 catches      = 0;
  static UInt16 change_views = 0;
  static UInt16 dashes       = 0;
  static UInt16 kicks        = 0;
  static UInt16 moves        = 0;
  static UInt16 says         = 0;
  static UInt16 turns        = 0;
  static UInt16 turn_necks   = 0;

  Int16 old_catches = catches;
  Int16 old_change_views = change_views;
  Int16 old_dashes = dashes;
  Int16 old_kicks = kicks;
  Int16 old_moves = moves;
  Int16 old_says = says;
  Int16 old_turns = turns;
  Int16 old_turn_necks = turn_necks;

  catches = sb_p->catchCount();
  change_views = sb_p->changeViewCount();
  dashes = sb_p->dashCount();
  kicks = sb_p->kickCount();
  moves = sb_p->moveCount();
  says = sb_p->sayCount();
  turns = sb_p->turnCount();
  turn_necks = sb_p->turnNeckCount();
    
  old_catches      = catches - old_catches;
  old_change_views = change_views - old_change_views;
  old_dashes       = dashes - old_dashes;
  old_kicks        = kicks - old_kicks;
  old_moves        = moves - old_moves;
  old_says         = says - old_says;
  old_turns        = turns - old_turns;
  old_turn_necks   = turn_necks - old_turn_necks;

  while (!sensors->commandQueue.empty()) {
    switch (sensors->commandQueue.front()) {
    case CMD_Catch:
      --old_catches;
      break;
    case CMD_ChangeView:
      --old_change_views;
      break;
    case CMD_Dash:
      --old_dashes;
      break;
    case CMD_Kick:
      --old_kicks;
      break;
    case CMD_Move:
      --old_moves;
      break;
    case CMD_Say:
      --old_says;
      break;
    case CMD_Turn:
      --old_turns;
      break;
    case CMD_TurnNeck:
      --old_turn_necks;
      break;
    default:
      //CMD_Unknown, CMD_Bye, CMD_Init, CMD_SenseBody
      break;
    }
    sensors->commandQueue.pop();
  }

  if ( old_catches < 0 ||
       old_change_views < 0 ||
       old_dashes < 0 ||
       old_kicks < 0 ||
       old_moves < 0 ||
       old_says < 0 ||
       old_turns < 0 ||
       old_turn_necks < 0 )
    sensors->theController->delayedActions(old_catches, old_change_views,
					   old_dashes, old_kicks, old_moves,
					   old_says, old_turns, old_turn_necks);

  if ( old_catches > 0 ||
       old_change_views > 0 ||
       old_dashes > 0 ||
       old_kicks > 0 ||
       old_moves > 0 ||
       old_says > 0 ||
       old_turns > 0 ||
       old_turn_necks > 0 )
    sensors->theController->missingActions(old_catches, old_change_views,
					   old_dashes, old_kicks, old_moves,
					   old_says, old_turns, old_turn_necks);

  ViewQuality view_quality = getViewQuality(sb_p->viewMode().quality());
  ViewWidth view_width = getViewWidth(sb_p->viewMode().width());
  Float32 head_ang = sb_p->headAngle().deg();
  Float32 speed_dist = sb_p->vel().dist();
  Float32 speed_dir = sb_p->vel().dir().deg();

  updateTime(sb_p->time());
  
  sensors->theSensorInterface->senseBody(view_quality, view_width, sb_p->stamina(),
					 sb_p->effort(), speed_dist, speed_dir,
					 head_ang, catches, change_views, dashes,
					 kicks, moves, says, turns, turn_necks);

  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Sensed!");
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_ServerParam>& sp_p)
{
  setLastSensorType(SENSOR_PARAM);  
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_Substitution>& sub_p)
{
  setLastSensorType(SENSOR_SUBSTITUTION);
}


CoordPoint<Float32U>
SensorDataHandler::getVectorUnknown(const RCC_Vector2d& vec) const
{
  /*
  RS::Float32U x;
  RS::Float32U y;
  if ( vec.isValid() ) {
    std::cerr << "Point(";
    if ( vec.dist() != HUGE_VAL && !isnan(vec.dist()) ) {
      std::cerr << vec.dist() << ", ";
      x = vec.dist();
    } else {
      std::cerr << "unknown, ";
    }
    if ( vec.dir().isValid() && !isnan(vec.dir().deg()) ) {
      std::cerr << vec.dir().deg() << ")";
      y = vec.dir().deg();
    } else {
      std::cerr << "unknown)";
    }
  } else {
    std::cerr << "UnknownPoint(";
  }

  RS::CoordPoint<RS::Float32U> pt(x,y);
  std::cerr << "\t" << pt << std::endl;
  return pt;
  */

  Float32U x;
  Float32U y;
  if ( vec.isValid() ) {
    if ( vec.dist() != HUGE_VAL && !isnan(vec.dist()) ) {
      x = vec.dist();
    } else {
      x.makeUnknown();
    }
    if ( vec.dir().isValid() && !isnan(vec.dir().deg()) ) {
      y = vec.dir().deg();
    } else {
      y.makeUnknown();
    }
  }
  
  return CoordPoint<Float32U>(x,y);  
}


CoordPoint<Float32U>
SensorDataHandler::getPointUnknown(const RCC_Vector2d& vec) const
{
  Float32U x;
  Float32U y;
  if ( vec.isValid() ) {
    if ( vec.x() != HUGE_VAL && !isnan(vec.x()) ) {
      x = vec.x();
    } else {
      x.makeUnknown();
    }
    if ( vec.y() != HUGE_VAL && !isnan(vec.y()) ) {
      y = vec.y();
    } else {
      y.makeUnknown();
    }
  }
  
  return CoordPoint<Float32U>(x,y);  
}


SideLineName
SensorDataHandler::getSideLineName(const RCC_VisualObject::location_t& loc) const
{
  SideLineName sline = SL_UNKNOWN;
  switch ( loc ) {
  case RCC_VisualObject::TOP_LOC:
    sline = SL_TOP;
    break;
  case RCC_VisualObject::BOTTOM_LOC:
    sline = SL_BOTTOM;
    break;
  case RCC_VisualObject::LEFT_LOC:
    sline = SL_LEFT;
    break;
  case RCC_VisualObject::RIGHT_LOC:
    sline = SL_RIGHT;
    break;
  default:
    RS_ERROR("Unknown location: " << loc);
  }
  return sline;
}


MarkerName
SensorDataHandler::getGoalName(const int l) const
{
  static const int GOAL_L_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::LEFT_LOC );
  static const int GOAL_R_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::RIGHT_LOC );

  MarkerName mn = UNKNOWN_GOAL;

  if ( l == GOAL_L_NUM ) {
    mn = GOAL_L;
  } else if ( l == GOAL_R_NUM ) {
    mn = GOAL_R;
  } else {
    mn = UNKNOWN_GOAL;
  }

  /*
  switch ( l ) {
  case GOAL_L_NUM:
    mn = GOAL_L;
    break;
  case GOAL_R_NUM:
    mn = GOAL_R;
    break;
  default:
    mn = UNKNOWN_GOAL;
  }
  */

  return mn;
}


MarkerName
SensorDataHandler::getFlagName(const int l) const
{
  /*
    INVALID_LOC,
    CLOSE_LOC,
    TOP_LOC, 
    BOTTOM_LOC, 
    LEFT_LOC, 
    RIGHT_LOC, 
    CENTER_LOC, 
    PENALTY_LOC, 
    GOAL_LOC,
    CLOSE_GOAL_LOC,
    MAX_LOC
  */
  static const int FLAG_C_NUM =
    RCC_Flag::getNum( RCC_VisualObject::CENTER_LOC );
  static const int FLAG_CT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::CENTER_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_CB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::CENTER_LOC,
		      RCC_VisualObject::BOTTOM_LOC );
  static const int FLAG_LT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::CENTER_LOC );
  static const int FLAG_LB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );
  static const int FLAG_RT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_RB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );

  static const int FLAG_PLT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_PLC_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::CENTER_LOC );
  static const int FLAG_PLB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );
  static const int FLAG_PRT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_PRC_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::CENTER_LOC );
  static const int FLAG_PRB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::PENALTY_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );

  static const int FLAG_GLT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_GLB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );
  static const int FLAG_GRT_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC );
  static const int FLAG_GRB_NUM =
    RCC_Flag::getNum( RCC_VisualObject::GOAL_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC );

  static const int FLAG_TL50_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      50 );
  static const int FLAG_TL40_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      40 );
  static const int FLAG_TL30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      30 );
  static const int FLAG_TL20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      20 );
  static const int FLAG_TL10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      10 );
  static const int FLAG_T0_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      0 );
  static const int FLAG_TR10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      10 );
  static const int FLAG_TR20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      20 );
  static const int FLAG_TR30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      30 );
  static const int FLAG_TR40_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      40 );
  static const int FLAG_TR50_NUM =
    RCC_Flag::getNum( RCC_VisualObject::TOP_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      50 );

  static const int FLAG_BL50_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      50 );
  static const int FLAG_BL40_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      40 );
  static const int FLAG_BL30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      30 );
  static const int FLAG_BL20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      20 );
  static const int FLAG_BL10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::LEFT_LOC,
		      10 );
  static const int FLAG_B0_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      0 );
  static const int FLAG_BR10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      10 );
  static const int FLAG_BR20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      20 );
  static const int FLAG_BR30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      30 );
  static const int FLAG_BR40_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      40 );
  static const int FLAG_BR50_NUM =
    RCC_Flag::getNum( RCC_VisualObject::BOTTOM_LOC,
		      RCC_VisualObject::RIGHT_LOC,
		      50 );

  static const int FLAG_LT30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      30 );
  static const int FLAG_LT20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      20 );
  static const int FLAG_LT10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      10 );
  static const int FLAG_L0_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      0 );
  static const int FLAG_LB10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      10 );
  static const int FLAG_LB20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      20 );
  static const int FLAG_LB30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::LEFT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      30 );

  static const int FLAG_RT30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      30 );
  static const int FLAG_RT20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      20 );
  static const int FLAG_RT10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::TOP_LOC,
		      10 );
  static const int FLAG_R0_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      0 );
  static const int FLAG_RB10_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      10 );
  static const int FLAG_RB20_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      20 );
  static const int FLAG_RB30_NUM =
    RCC_Flag::getNum( RCC_VisualObject::RIGHT_LOC,
		      RCC_VisualObject::BOTTOM_LOC,
		      30 );

  MarkerName mn;

  /*
  switch ( l ) {
  case FLAG_C_NUM:
    mn = FLAG_C;
    break;
  case FLAG_CT_NUM:
    mn = FLAG_CT;
    break;
  case FLAG_CB_NUM:
    mn = FLAG_CB;
    break;
  case FLAG_LT_NUM:
    mn = FLAG_LT;
    break;
  case FLAG_LB_NUM:
    mn = FLAG_LB;
    break;
  case FLAG_RT_NUM:
    mn = FLAG_RT;
    break;
  case FLAG_RB_NUM:
    mn = FLAG_RB;
    break;
  case FLAG_PLT_NUM:
    mn = FLAG_PLT;
    break;
  case FLAG_PLC_NUM:
    mn = FLAG_PLC;
    break;
  case FLAG_PLB_NUM:
    mn = FLAG_PLB;
    break;
  case FLAG_PRT_NUM:
    mn = FLAG_PRT;
    break;
  case FLAG_PRC_NUM:
    mn = FLAG_PRC;
    break;
  case FLAG_PRB_NUM:
    mn = FLAG_PRB;
    break;
  case FLAG_GLT_NUM:
    mn = FLAG_GLT;
    break;
  case FLAG_GLB_NUM:
    mn = FLAG_GLB;
    break;
  case FLAG_GRT_NUM:
    mn = FLAG_GRT;
    break;
  case FLAG_GRB_NUM:
    mn = FLAG_GRB;
    break;
  case FLAG_TL50_NUM:
    mn = FLAG_TL50;
    break;
  case FLAG_TL40_NUM:
    mn = FLAG_TL40;
    break;
  case FLAG_TL30_NUM:
    mn = FLAG_TL30;
    break;
  case FLAG_TL20_NUM:
    mn = FLAG_TL20;
    break;
  case FLAG_TL10_NUM:
    mn = FLAG_TL10;
    break;
  case FLAG_T0_NUM:
    mn = FLAG_T0;
    break;
  case FLAG_TR10_NUM:
    mn = FLAG_TR10;
    break;
  case FLAG_TR20_NUM:
    mn = FLAG_TR20;
    break;
  case FLAG_TR30_NUM:
    mn = FLAG_TR30;
    break;
  case FLAG_TR40_NUM:
    mn = FLAG_TR40;
    break;
  case FLAG_TR50_NUM:
    mn = FLAG_TR50;
    break;
  case FLAG_BL50_NUM:
    mn = FLAG_BL50;
    break;
  case FLAG_BL40_NUM:
    mn = FLAG_BL40;
    break;
  case FLAG_BL30_NUM:
    mn = FLAG_BL30;
    break;
  case FLAG_BL20_NUM:
    mn = FLAG_BL20;
    break;
  case FLAG_BL10_NUM:
    mn = FLAG_BL10;
    break;
  case FLAG_B0_NUM:
    mn = FLAG_B0;
    break;
  case FLAG_BR10_NUM:
    mn = FLAG_BR10;
    break;
  case FLAG_BR20_NUM:
    mn = FLAG_BR20;
    break;
  case FLAG_BR30_NUM:
    mn = FLAG_BR30;
    break;
  case FLAG_BR40_NUM:
    mn = FLAG_BR40;
    break;
  case FLAG_BR50_NUM:
    mn = FLAG_BR50;
    break;
  case FLAG_LT30_NUM:
    mn = FLAG_LT30;
    break;
  case FLAG_LT20_NUM:
    mn = FLAG_LT20;
    break;
  case FLAG_LT10_NUM:
    mn = FLAG_LT10;
    break;
  case FLAG_L0_NUM:
    mn = FLAG_L0;
    break;
  case FLAG_LB10_NUM:
    mn = FLAG_LB10;
    break;
  case FLAG_LB20_NUM:
    mn = FLAG_LB20;
    break;
  case FLAG_LB30_NUM:
    mn = FLAG_LB30;
    break;
  case FLAG_RT30_NUM:
    mn = FLAG_RT30;
    break;
  case FLAG_RT20_NUM:
    mn = FLAG_RT20;
    break;
  case FLAG_RT10_NUM:
    mn = FLAG_RT10;
    break;
  case FLAG_R0_NUM:
    mn = FLAG_R0;
    break;
  case FLAG_RB10_NUM:
    mn = FLAG_RB10;
    break;
  case FLAG_RB20_NUM:
    mn = FLAG_RB20;
    break;
  case FLAG_RB30_NUM:
    mn = FLAG_RB30;
    break;
  default:
    mn = UNKNOWN_FLAG;
  }
  */
  
  if ( l == FLAG_C_NUM ) {
    mn = FLAG_C;
  } else if ( l == FLAG_CT_NUM ) {
    mn = FLAG_CT;
  } else if ( l == FLAG_CB_NUM ) {
    mn = FLAG_CB;
  } else if ( l == FLAG_LT_NUM ) {
    mn = FLAG_LT;
  } else if ( l == FLAG_LB_NUM ) {
    mn = FLAG_LB;
  } else if ( l == FLAG_RT_NUM ) {
    mn = FLAG_RT;
  } else if ( l == FLAG_RB_NUM ) {
    mn = FLAG_RB;
  } else if ( l == FLAG_PLT_NUM ) {
    mn = FLAG_PLT;
  } else if ( l == FLAG_PLC_NUM ) {
    mn = FLAG_PLC;
  } else if ( l == FLAG_PLB_NUM ) {
    mn = FLAG_PLB;
  } else if ( l == FLAG_PRT_NUM ) {
    mn = FLAG_PRT;
  } else if ( l == FLAG_PRC_NUM ) {
    mn = FLAG_PRC;
  } else if ( l == FLAG_PRB_NUM ) {
    mn = FLAG_PRB;
  } else if ( l == FLAG_GLT_NUM ) {
    mn = FLAG_GLT;
  } else if ( l == FLAG_GLB_NUM ) {
    mn = FLAG_GLB;
  } else if ( l == FLAG_GRT_NUM ) {
    mn = FLAG_GRT;
  } else if ( l == FLAG_GRB_NUM ) {
    mn = FLAG_GRB;
  } else if ( l == FLAG_TL50_NUM ) {
    mn = FLAG_TL50;
  } else if ( l == FLAG_TL40_NUM ) {
    mn = FLAG_TL40;
  } else if ( l == FLAG_TL30_NUM ) {
    mn = FLAG_TL30;
  } else if ( l == FLAG_TL20_NUM ) {
    mn = FLAG_TL20;
  } else if ( l == FLAG_TL10_NUM ) {
    mn = FLAG_TL10;
  } else if ( l == FLAG_T0_NUM ) {
    mn = FLAG_T0;
  } else if ( l == FLAG_TR10_NUM ) {
    mn = FLAG_TR10;
  } else if ( l == FLAG_TR20_NUM ) {
    mn = FLAG_TR20;
  } else if ( l == FLAG_TR30_NUM ) {
    mn = FLAG_TR30;
  } else if ( l == FLAG_TR40_NUM ) {
    mn = FLAG_TR40;
  } else if ( l == FLAG_TR50_NUM ) {
    mn = FLAG_TR50;
  } else if ( l == FLAG_BL50_NUM ) {
    mn = FLAG_BL50;
  } else if ( l == FLAG_BL40_NUM ) {
    mn = FLAG_BL40;
  } else if ( l == FLAG_BL30_NUM ) {
    mn = FLAG_BL30;
  } else if ( l == FLAG_BL20_NUM ) {
    mn = FLAG_BL20;
  } else if ( l == FLAG_BL10_NUM ) {
    mn = FLAG_BL10;
  } else if ( l == FLAG_B0_NUM ) {
    mn = FLAG_B0;
  } else if ( l == FLAG_BR10_NUM ) {
    mn = FLAG_BR10;
  } else if ( l == FLAG_BR20_NUM ) {
    mn = FLAG_BR20;
  } else if ( l == FLAG_BR30_NUM ) {
    mn = FLAG_BR30;
  } else if ( l == FLAG_BR40_NUM ) {
    mn = FLAG_BR40;
  } else if ( l == FLAG_BR50_NUM ) {
    mn = FLAG_BR50;
  } else if ( l == FLAG_LT30_NUM ) {
    mn = FLAG_LT30;
  } else if ( l == FLAG_LT20_NUM ) {
    mn = FLAG_LT20;
  } else if ( l == FLAG_LT10_NUM ) {
    mn = FLAG_LT10;
  } else if ( l == FLAG_L0_NUM ) {
    mn = FLAG_L0;
  } else if ( l == FLAG_LB10_NUM ) {
    mn = FLAG_LB10;
  } else if ( l == FLAG_LB20_NUM ) {
    mn = FLAG_LB20;
  } else if ( l == FLAG_LB30_NUM ) {
    mn = FLAG_LB30;
  } else if ( l == FLAG_RT30_NUM ) {
    mn = FLAG_RT30;
  } else if ( l == FLAG_RT20_NUM ) {
    mn = FLAG_RT20;
  } else if ( l == FLAG_RT10_NUM ) {
    mn = FLAG_RT10;
  } else if ( l == FLAG_R0_NUM ) {
    mn = FLAG_R0;
  } else if ( l == FLAG_RB10_NUM ) {
    mn = FLAG_RB10;
  } else if ( l == FLAG_RB20_NUM ) {
    mn = FLAG_RB20;
  } else if ( l == FLAG_RB30_NUM ) {
    mn = FLAG_RB30;
  } else {
    mn = UNKNOWN_FLAG;
  }

  return mn;
}


void
SensorDataHandler::handle(const RCC_Pointer<RCC_Visual>& pt_p)
{
  setLastSensorType(SENSOR_VISUAL);
  updateTime(pt_p->time());
  for ( unsigned int i = 0; i < pt_p->numOfBalls(); ++i ) {
    const RCC_Pointer < RCC_Ball >& ball = pt_p->ball(i);
    CoordPoint<Float32U> pos = getVectorUnknown(ball->relPos());
    //RS_LOG ( LA_SENSORS, toString(sensors->getCurrentTime()) << "pos = " << pos );
    // ball->relVel(); // Precalculated relative speed!
    //std::cerr << *ball << std::endl;
    CoordPoint<Float32U> ddchg;
    if ( ball->relVel().isValid () ) {
      ddchg = getPointUnknown(ball->distDirChg());
      //ddchg = getVectorUnknown(ball->relVel());
    }
    //RS_LOG ( LA_SENSORS, toString(sensors->getCurrentTime()) << "ddchg = " << ddchg );
    CoordPoint<Float32U> speed = getVectorUnknown(ball->relVel());
    sensors->theSensorInterface->seeBall(pos.getX(),
					 pos.getY(),
					 ddchg.getX(),
					 ddchg.getY(),
					 speed.getX(),
					 speed.getY());
  }

  for ( unsigned int i = 0; i < pt_p->numOfLines(); ++i ) {
    const RCC_Pointer < RCC_Line >& line = pt_p->line(i);
    CoordPoint<Float32U> pos = getVectorUnknown(line->relPos());
    sensors->theSensorInterface->seeLine(getSideLineName(line->name()),
					 pos.getX(),
					 pos.getY());
  }

  for ( unsigned int i = 0; i < pt_p->numOfFlags(); ++i ) {
    const RCC_Pointer < RCC_Flag >& flag = pt_p->flag(i);
    CoordPoint<Float32U> pos = getVectorUnknown(flag->relPos());
    sensors->theSensorInterface->seeMarker(getFlagName(flag->name()),
					   pos.getX(),
					   pos.getY(),
					   flag->name() == RCC_Flag::getNum( RCC_VisualObject::CLOSE_LOC ));
  }
  
  for ( unsigned int i = 0; i < pt_p->numOfGoals(); ++i ) {
    const RCC_Pointer < RCC_Flag >& goal = pt_p->goal(i);
    CoordPoint<Float32U> pos = getVectorUnknown(goal->relPos());
    sensors->theSensorInterface->seeMarker(getGoalName(goal->name()),
					   pos.getX(),
					   pos.getY(),
					   goal->name() == RCC_Flag::getNum( RCC_VisualObject::CLOSE_GOAL_LOC ));
  }

  /*
  for ( unsigned int i = 0; i < pt_p->numOfAllies(); ++i ) {
    RS_LOG( LA_SENSORS, "Saw a teammate!" );
    const RCC_Pointer < RCC_Player >& player = pt_p->ally(i);
    RCC_PlayerName pname = player->playerName();
    Side side = sensors->theSensorInterface->getOurSide();
    Team team = T_OUR_TEAM;
    UInt16U unum = pname.uNum();
    if ( unum == RCC_PlayerName::UNKNOWN_NUM ) {
      unum.makeUnknown();
    }
    CoordPoint<Float32U> pos = getVectorUnknown(player->relPos());
    CoordPoint<Float32U> ddchg;
    if ( player->relVel().isValid() ) {
      ddchg = getPointUnknown(player->distDirChg());
    }
    Float32U bodyDir = player->bodyDir().deg();
    if ( bodyDir == HUGE_VAL ) {
      bodyDir.makeUnknown();
    }
    Float32U headDir = player->headDir().deg();
    if ( headDir == HUGE_VAL ) {
      headDir.makeUnknown();
    }
    sensors->theSensorInterface->seePlayer(side,
					   team,
					   unum,
					   pos.getX(),
					   pos.getY(),
					   ddchg.getX(),
					   ddchg.getY(),
					   bodyDir,
					   headDir);
  }
  for ( unsigned int i = 0; i < pt_p->numOfNeuts(); ++i ) {
    RS_LOG( LA_SENSORS, "Saw an unknown player!" );
    const RCC_Pointer < RCC_Player >& player = pt_p->neut(i);
    RCC_PlayerName pname = player->playerName();
    Side side = S_UNKNOWN;
    Team team = T_UNKNOWN;
    UInt16U unum = pname.uNum();
    if ( unum == RCC_PlayerName::UNKNOWN_NUM ) {
      unum.makeUnknown();
    }
    CoordPoint<Float32U> pos = getVectorUnknown(player->relPos());
    CoordPoint<Float32U> ddchg;
    if ( player->relVel().isValid() ) {
      ddchg = getPointUnknown(player->distDirChg());
    }
    Float32U bodyDir = player->bodyDir().deg();
    if ( bodyDir == HUGE_VAL ) {
      bodyDir.makeUnknown();
    }
    Float32U headDir = player->headDir().deg();
    if ( headDir == HUGE_VAL ) {
      headDir.makeUnknown();
    }
    sensors->theSensorInterface->seePlayer(side,
					   team,
					   unum,
					   pos.getX(),
					   pos.getY(),
					   ddchg.getX(),
					   ddchg.getY(),
					   bodyDir,
					   headDir);
  }
  for ( unsigned int i = 0; i < pt_p->numOfEnemies(); ++i ) {
    RS_LOG( LA_SENSORS, "Saw an opponent!" );
    const RCC_Pointer < RCC_Player >& player = pt_p->enemy(i);
    RCC_PlayerName pname = player->playerName();
    Side side = sensors->theSensorInterface->getTheirSide();
    Team team = T_THEIR_TEAM;
    UInt16U unum = pname.uNum();
    if ( unum == RCC_PlayerName::UNKNOWN_NUM ) {
      unum.makeUnknown();
    }
    CoordPoint<Float32U> pos = getVectorUnknown(player->relPos());
    CoordPoint<Float32U> ddchg;
    if ( player->relVel().isValid() ) {
      ddchg = getPointUnknown(player->distDirChg());
    }
    Float32U bodyDir = player->bodyDir().deg();
    if ( bodyDir == HUGE_VAL ) {
      bodyDir.makeUnknown();
    }
    Float32U headDir = player->headDir().deg();
    if ( headDir == HUGE_VAL ) {
      headDir.makeUnknown();
    }
    sensors->theSensorInterface->seePlayer(side,
					   team,
					   unum,
					   pos.getX(),
					   pos.getY(),
					   ddchg.getX(),
					   ddchg.getY(),
					   bodyDir,
					   headDir);
  }
  */
  for ( unsigned int i = 0; i < pt_p->numOfPlayers(); ++i ) {
    const RCC_Pointer < RCC_Player >& player = pt_p->player(i);
    RCC_PlayerName pname = player->playerName();
    Side side = S_UNKNOWN;
    Team team = T_UNKNOWN;
    if ( pname.isKnown() ) {
      if ( !strcmp(pname.team().c_str(), sensors->theSensorInterface->getOurTeamName().c_str()) ) {
	RS_LOG( LA_SENSORS, "Saw a teammate!" );
	team = T_OUR_TEAM;
	side = sensors->theSensorInterface->getOurSide();
      } else {
	RS_LOG( LA_SENSORS, "Saw an opponent!" );
	team = T_THEIR_TEAM;
	side = sensors->theSensorInterface->getTheirSide();
      } 
    } else {
      RS_LOG( LA_SENSORS, "Saw an unknown player!" );
    }
    UInt16U unum = pname.uNum();
    if ( unum == RCC_PlayerName::UNKNOWN_NUM ) {
      unum.makeUnknown();
    }
    CoordPoint<Float32U> pos = getVectorUnknown(player->relPos());
    CoordPoint<Float32U> ddchg;
    if ( player->relVel().isValid() ) {
      ddchg = getPointUnknown(player->distDirChg());
    }
    Float32U bodyDir = player->bodyDir().deg();
    if ( bodyDir == HUGE_VAL ) {
      bodyDir.makeUnknown();
    }
    Float32U headDir = player->headDir().deg();
    if ( headDir == HUGE_VAL ) {
      headDir.makeUnknown();
    }
    sensors->theSensorInterface->seePlayer(side,
					   team,
					   unum,
					   pos.getX(),
					   pos.getY(),
					   ddchg.getX(),
					   ddchg.getY(),
					   bodyDir,
					   headDir);
  }

  RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + ": Saw!");
}


void
SensorDataHandler::updateTime(int time)
{
  sensors->setCurrentTime(time);
  if ( sensors->theController->updateTime(sensors->getCurrentTime()) ) {
    sensors->theSensorInterface->newCycle();
    RS_LOG( LA_SENSORS, toString(sensors->getCurrentTime()) + "--- NEW CYCLE ---");
  }
}


void
SensorDataHandler::setLastSensorType(SensorType sensor_type)
  throw()
{
  sensors->setLastSensorType(sensor_type);
}


Sensors::Sensors(SensorInterface* sensor_interface)
  : theController(NULL),
    theSensorInterface(sensor_interface),
    theParser(RCC_Parser::FIELD_PLAYER, sensor_interface->getOurTeamName()),
    sensorDataHandler(this),
    lastSensorType(SENSOR_UNKNOWN),
    currentTime(0),
    commandQueue()
{
  SensorDataHandler* sdh = &sensorDataHandler;
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_SenseBody > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_ServerParam > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_PlayerParam > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_PlayerType > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_Visual > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_RefAudio > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_SelfAudio > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_PlayerAudio > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_CoachAudio > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_Score > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_InitMsg > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_Error > >*)sdh);
  theParser.registerHandler(*(RCC_Handler < RCC_Pointer < RCC_Substitution > >*)sdh);
}


void
Sensors::init(Controller* controller)
{
  theController = controller;
}


void
Sensors::analyze(ServerCommand* command)
{
  lastSensorType = SENSOR_ACTUATORS;
  commandQueue.push(command->getType());
  theSensorInterface->setLastCommand(command);
  theSensorInterface->setLastSensorType(lastSensorType);  
}


void
Sensors::analyze(char* sensor_data)
{
  lastSensorType = SENSOR_UNKNOWN;
  char* sd = sensor_data;
  RS_BEGIN_TRY
    theParser.parse(sensor_data);
  RS_END_TRY
    RS_CATCH( ...,
	      RS_ERROR( "Caught an exception from RCCParser!" ); )

  if ( lastSensorType == SENSOR_UNKNOWN ) {
    RS_LOG( LA_SENSORS, "UNKNOWN: " << sd );
  }
  theSensorInterface->setLastSensorType(lastSensorType);
}


#if 0
void
Sensors::analyzeVisualSensor(char* sightInfo)
{

  RS_LOG( LA_SENSORS, "SAW: " << sightInfo );
  
  static ObjectType object_type;  
  static ViewQuality view_qual;

  static Float32U dist;
  static Float32U ang;
  static Float32U dir_chng;
  static Float32U dist_chng;
  static Float32U body_dir;
  static Float32U face_dir;
  static UInt16U player_number;
  static char player_team_name[64];
  static Side player_side;
  static Team player_team;
  static MarkerName marker;
  static SideLineName line;
  static bool behind;

  while (*sightInfo != ')') {

    dist.makeUnknown();
    dir_chng.makeUnknown();
    dist_chng.makeUnknown();
    body_dir.makeUnknown();
    face_dir.makeUnknown();
    player_number.makeUnknown();
    player_team_name[0] = '\0';
    player_side = S_UNKNOWN;
    player_team = T_UNKNOWN;
    marker = UNKNOWN_MARKER;
    line = SL_UNKNOWN;
    behind = false;
 
    //std::cerr << sightInfo << std::endl; //"Before getWord...\n";
    sightInfo += 3;          // " ((" 
    //getWord(&sightInfo);            // " ((" 
    //std::cerr << sightInfo << std::endl;
    //std::cerr << "After getWord...\n";
    //std::cerr << ".";
    
    if ( *sightInfo=='g' ) {
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "goal " 
      if ( *sightInfo=='r' )
	marker = GOAL_R;
      else if ( *sightInfo=='l' )
	marker = GOAL_L;
      else
	RS_ERROR("goal ?");
    } else if ( *sightInfo=='G' ){
      object_type = OBJ_MARKER_BEHIND;
      behind = true;
      marker = UNKNOWN_GOAL;
    } else if ( *sightInfo=='f' ){
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "flag " 
      if ( *sightInfo=='r' ){
	sightInfo+=2;
	if ( *sightInfo=='0' )
	  marker = FLAG_R0;
	else if ( *sightInfo=='b' ){
	  sightInfo+=1;
	  if ( *sightInfo==')' )
	    marker = FLAG_RB; 
	  else {
	    sightInfo+=1;
	    if ( *sightInfo=='1' )   marker = FLAG_RB10; else
	      if ( *sightInfo=='2' )   marker = FLAG_RB20; else
		if ( *sightInfo=='3' )   marker = FLAG_RB30; else
		  RS_ERROR("flag r b ?");
	  }  
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=1;
	    if ( *sightInfo==')' )     marker = FLAG_RT; 
	    else{
	      sightInfo+=1;
	      if ( *sightInfo=='1' )   marker = FLAG_RT10; else
		if ( *sightInfo=='2' )   marker = FLAG_RT20; else
		  if ( *sightInfo=='3' )   marker = FLAG_RT30; else
		    RS_ERROR("flag r t ?");
	    }  
	  } else
	    RS_ERROR("flag r ?");
      } else
	if ( *sightInfo=='l' ){
	  sightInfo+=2;
	  if ( *sightInfo=='0' )       marker = FLAG_L0;   else
	    if ( *sightInfo=='b' ){
	      sightInfo+=1;
	      if ( *sightInfo==')' )     marker = FLAG_LB; 
	      else{
		sightInfo+=1;
		if ( *sightInfo=='1' )   marker = FLAG_LB10; else
		  if ( *sightInfo=='2' )   marker = FLAG_LB20; else
		    if ( *sightInfo=='3' )   marker = FLAG_LB30; else
		      RS_ERROR("flag l b ?");
	      }  
	    } else
	      if ( *sightInfo=='t' ){
		sightInfo+=1;
		if ( *sightInfo==')' )     marker = FLAG_LT; 
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='1' )   marker = FLAG_LT10; else
		    if ( *sightInfo=='2' )   marker = FLAG_LT20; else
		      if ( *sightInfo=='3' )   marker = FLAG_LT30; else
			RS_ERROR("flag l t ?");
		}  
	      } else
		RS_ERROR("flag l ?");
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=2;
	    if ( *sightInfo=='0' )       marker = FLAG_T0;   else
	      if ( *sightInfo=='l' ){
		sightInfo+=2;
		if ( *sightInfo=='1' )     marker = FLAG_TL10; else
		  if ( *sightInfo=='2' )     marker = FLAG_TL20; else
		    if ( *sightInfo=='3' )     marker = FLAG_TL30; else
		      if ( *sightInfo=='4' )     marker = FLAG_TL40; else
			if ( *sightInfo=='5' )     marker = FLAG_TL50; else
			  RS_ERROR("flag t l ?");
	      } else
		if ( *sightInfo=='r' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_TR10; else
		    if ( *sightInfo=='2' )     marker = FLAG_TR20; else
		      if ( *sightInfo=='3' )     marker = FLAG_TR30; else
			if ( *sightInfo=='4' )     marker = FLAG_TR40; else
			  if ( *sightInfo=='5' )     marker = FLAG_TR50; else
			    RS_ERROR("flag t r ?");
		} else
		  RS_ERROR("flag t ?");
	  } else
	    if ( *sightInfo=='b' ){
	      sightInfo+=2;
	      if ( *sightInfo=='0' )       marker = FLAG_B0;   else
		if ( *sightInfo=='l' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_BL10; else
		    if ( *sightInfo=='2' )     marker = FLAG_BL20; else
		      if ( *sightInfo=='3' )     marker = FLAG_BL30; else
			if ( *sightInfo=='4' )     marker = FLAG_BL40; else
			  if ( *sightInfo=='5' )     marker = FLAG_BL50; else
			    RS_ERROR("flag b l ?");
		} else
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='1' )     marker = FLAG_BR10; else
		      if ( *sightInfo=='2' )     marker = FLAG_BR20; else
			if ( *sightInfo=='3' )     marker = FLAG_BR30; else
			  if ( *sightInfo=='4' )     marker = FLAG_BR40; else
			    if ( *sightInfo=='5' )     marker = FLAG_BR50; else
			      RS_ERROR("flag b r ?");
		  } else
		    RS_ERROR("flag b ?");
	    } else
	      if ( *sightInfo=='c' ){
		sightInfo+=1;
		if ( *sightInfo==')' )       marker = FLAG_C;
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='b' )     marker = FLAG_CB; else
		    if ( *sightInfo=='t' )     marker = FLAG_CT; else
		      RS_ERROR("flag c ?");
		}
	      } else
		if ( *sightInfo=='p' ){
		  sightInfo+=2;
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='t')      marker = FLAG_PRT; else
		      if ( *sightInfo=='c')      marker = FLAG_PRC; else
			if ( *sightInfo=='b')      marker = FLAG_PRB; else	    
			  RS_ERROR("flag p r ?");
		  } else
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t')      marker = FLAG_PLT; else
			if ( *sightInfo=='c')      marker = FLAG_PLC; else
			  if ( *sightInfo=='b')      marker = FLAG_PLB; else	    
			    RS_ERROR("flag p l ?");
		    } else
		      RS_ERROR("flag p ?");
		} else
		  if ( *sightInfo=='g' ){
		    sightInfo+=2;
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t' )     marker = FLAG_GLT; else
			if ( *sightInfo=='b' )     marker = FLAG_GLB; else
			  RS_ERROR("flag g l ?");
		    } else
		      if ( *sightInfo=='r' ){
			sightInfo+=2;
			if ( *sightInfo=='t' )     marker = FLAG_GRT; else
			  if ( *sightInfo=='b' )     marker = FLAG_GRB; else
			    RS_ERROR("flag g r ?");
		      } else
			RS_ERROR("flag g ?");
		  } else
		    RS_ERROR("flag ?");
    } else
      if ( *sightInfo=='F' ){
	object_type = OBJ_MARKER_BEHIND;
	behind = true;
	marker = UNKNOWN_FLAG;
      } else
	if ( *sightInfo=='l' ){
	  object_type = OBJ_LINE;
	  sightInfo+=5;          // "line " 
	  if ( *sightInfo=='r' )         line   = SL_RIGHT;  else
	    if ( *sightInfo=='l' )         line   = SL_LEFT;   else
	      if ( *sightInfo=='t' )         line   = SL_TOP;    else
		if ( *sightInfo=='b' )         line   = SL_BOTTOM; else
		  RS_ERROR("line ?");
	} else
	  if ( *sightInfo=='p' || *sightInfo=='P' ){
	    object_type = OBJ_PLAYER;
	    sightInfo+=6;                          // "player" 
	    if ( *sightInfo == ' ' ){              // there's a team  
	      sightInfo++;
	      // get the team name
	      int a = 0;
	      while ( *sightInfo != ' ' && *sightInfo != ')' && a < 63 )
		player_team_name[a++]=*sightInfo++;
	      if ( a == 64 )
		RS_WARNING("team name WAY to long > 63 characters!");
	      player_team_name[a] = '\0';

	      if ( !strcmp(player_team_name, theSensorInterface->getOurTeamName().c_str()) ) {
		player_side = theSensorInterface->getOurSide();
		player_team = T_OUR_TEAM;
	      } else if ( theSensorInterface->getTheirTeamName() == "" ) {
		player_side = theSensorInterface->getTheirSide();
		player_team = T_THEIR_TEAM;
		theSensorInterface->setTheirTeamName(player_team_name);
	      } else if ( !strcmp(player_team_name,
				  theSensorInterface->getTheirTeamName().c_str()) ) {
		player_side = theSensorInterface->getTheirSide();
		player_team = T_THEIR_TEAM;
	      }
	
	      if ( *sightInfo== ' ' ){             // there's a number 
		player_number = getInt(&sightInfo);
	      }
	    }
	  } else
	    if ( *sightInfo=='b' || *sightInfo=='B' ) {
	      object_type = OBJ_BALL;
	    }
	    else {
	      RS_ERROR("unknown object in analyzeVisualSensors: " + std::string(sightInfo));
	      return;
	    }
  
    advanceTo(')',&sightInfo);              // advance to end of object 


    ang = getFloat(&sightInfo);

    if ( *sightInfo != ')' ) {                  // 'high' quality     
      view_qual = VQ_HIGH;
      dist = ang;
      ang = getFloat(&sightInfo);
    }
    else {
      view_qual = VQ_LOW;
    }

    if ( *sightInfo != ')' ){
      dist_chng = getFloat(&sightInfo);
      dir_chng  = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ){
      if (object_type != OBJ_PLAYER)
	RS_WARNING("Only players should have body_dir");
      body_dir = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ){
      if (object_type != OBJ_PLAYER)
	RS_WARNING("Only players should have face_dir");
      face_dir = getFloat(&sightInfo);
    }

    if ( *sightInfo != ')' ) {
      RS_ERROR("Should be done with object info here");
      return;
    }

    sightInfo++;                                // ")" 


    //std::cerr << "Before switch...\n";
    
    switch (object_type){
    case OBJ_MARKER_BEHIND:
    case OBJ_MARKER:
      theSensorInterface->seeMarker(marker, dist, ang, behind);
      //theSensorInterface->seeMarker(marker, dist, ang, dist_chng, dir_chng, behind);
      break;
    case OBJ_LINE:
      theSensorInterface->seeLine(line, dist, ang);
      break;
    case OBJ_BALL:
      theSensorInterface->seeBall(dist, ang, dist_chng, dir_chng);
      break;
    case OBJ_PLAYER:
      /*
	if ( !player_side ){                      // Too far for team or num 
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(ang, time);
	else if ( dir_chng.isUnknown() )                   // high quality  
	seePlayer(dist, ang, time);
	else {                                        // know direction
	RS_ERROR("Shouldn't know dir_chng when the player's far");
	return;
	}
	}  

	else{
	if ( !player_number ){                  // Too far for number     
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(player_side, ang, time);
	else if ( dir_chng.isUnknown() )                   // high quality  
	seePlayer(player_side, dist, ang, time);
	else {                                           // know direction
	RS_ERROR("Shouldn't know dir_chng when the team member's far");
	return;
	}
	}

	else{                                   // Know side AND number   
	if ( view_qual == VQ_LOW )                // low quality   
	seePlayer(player_side, player_number, ang, time);
	else if ( dir_chng.IsUnknown() ){                  // high quality  
	printf("%s\n",sightInfo-30);
	RS_ERROR("Should know dir_chng when know number");
	seePlayer(player_side, player_number, dist, ang, time);
	}
	else */                                       // know direction

      //cout << "Saw player " << player_side << ":" << player_number << endl;
      
      theSensorInterface->seePlayer(player_side, player_team, player_number,
				    dist, ang, dist_chng, dir_chng,
				    body_dir, face_dir);
      /*
	}
	}
	*/
    }
    //std::cerr << "sightInfo = " << sightInfo << std::endl;
  }
  //std::cerr << "sightInfo = " << sightInfo << std::endl;
}


/****************************************************************************************/

/*
void
Sensors::analyzeAuralSensor(char* sound_info)
{
  if ( sound_info[1] == 'r' ){   // Referee or Coach message 
    sound_info += 9;             // " referee "
    // I'm not sure this is correct after the last changes!!!
    if ( islower(sound_info[0]) )
      analyzeRefereeSound(sound_info);  // Coach messages start with capital letters or are numbers
    else
      // Return the rest of the string, even the last ')'
      theSensorInterface->hearCoach(sound_info);
    return;
  }
  
  // Return the rest of the string, even the last ')'
  theSensorInterface->hearPlayer(sound_info);
  
}
*/

/****************************************************************************************/


void
Sensors::analyzeRefereeSound(char* msg)
{

  static const char left_side = 'l';
  static const char right_side = 'r';
  
  PlayMode play_mode = PM_UNKNOWN;
  Side side = S_UNKNOWN;
  
  switch ( msg[0] ) {
  case 'p': play_mode = PM_PLAY_ON; break;           // play_on
  case 'k': 
    if ( msg[5] == 'i' ){                                  // kick_in
      play_mode = PM_KICK_IN;
      if ( msg[8] == left_side )
	side = S_LEFT;
      else if ( msg[8] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("kick_in_?");
    }
    else if ( msg[5] == 'o' ){                            // kick_off
      play_mode = PM_KICK_OFF;
      if ( msg[9] == left_side )
	side = S_LEFT;
      else if ( msg[9] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("kick_off_?");
    }
    else
      RS_ERROR("referee k..?");
    break;
  case 'g': 
    if ( msg[5] == 'k' ) {                                 // goal_kick 
      play_mode = PM_GOAL_KICK;
      if ( msg[10] == left_side )
	side = S_LEFT;
      else if ( msg[10] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("goal_kick_?");
    }
    else if ( msg[5] == 'e' ){                           // goalie_catch_ball
      play_mode = PM_GOALIE_GOT_BALL;
      if ( msg[18] == left_side )
	side = S_LEFT;
      else if ( msg[18] == right_side )
	side = S_RIGHT;
      else
	RS_ERROR("goalie_catch_ball_?");
    }
    else if ( msg[5] == left_side ){                    // goal
      play_mode = PM_GOAL;
      side = S_LEFT;
      theSensorInterface->hearScore(side, getInt(&msg[7]));
    }
    else if ( msg[5] == right_side ){
      play_mode = PM_GOAL;
      side = S_RIGHT;
      theSensorInterface->hearScore(side, getInt(&msg[7]));
    }
    else 
      RS_ERROR("referee g..?");
    break;
  case 'c':                                               // corner_kick 
    play_mode = PM_CORNER_KICK;
    if ( msg[12] == left_side )
      side = S_LEFT;
    else if ( msg[12] == right_side )
      side = S_RIGHT;
    else 
      RS_ERROR("corner_kick_?");
    break;
  case 'd': play_mode = PM_DROP_BALL; break;        // drop_ball     
  case 'o':                                               // offside     
    play_mode = PM_OFFSIDE_KICK;
    if ( msg[8] == left_side )
      side = S_LEFT;
    else if ( msg[8] == right_side )
      side = S_RIGHT;
    else 
      RS_ERROR("offside_?");
    break;
  case 'f':
    if ( msg[5] == 'k' ){                                 // free_kick
      play_mode = PM_FREE_KICK;
      if ( msg[10] == left_side )
	side = S_LEFT;
      else if ( msg[10] == right_side )
	side = S_RIGHT;
      else 
	RS_ERROR("free_kick_?");
    }
    // ???? Why don't I have to do anything?
    else if ( msg[5] == left_side )                     // foul 
      ;
    else if ( msg[5] == right_side )
      ;
    else 
      RS_ERROR("referee f..?");
    break;
  case 'h':                                               // half_time 
    play_mode = PM_HALF_TIME;
    break;             
  case 'b':
    play_mode = PM_BEFORE_KICK_OFF;
    break;       // before_kick_off 
  case 't': 
    if ( msg[5] == 'u' ) {                             // time_up 
      play_mode = PM_TIME_UP; 
    }
    else if ( msg[5] == 'o' ) {                         // time_over 
      play_mode = PM_TIME_OVER;
      break;
    }
    else if ( msg[5] == 'e' ){                            // time_extended 
      play_mode = PM_EXTENDED_TIME;
    }
    else 
      RS_ERROR("referee t..?");
    break;
  default:
    RS_ERROR("Referee msg ????");
  }

  theSensorInterface->hearReferee(play_mode, side);
}

#endif


void
Sensors::analyzeCoachMessage(char* sensor_data)
{
  lastSensorType = SENSOR_COACH;
  char* sd = sensor_data;
  if ( sd[1] == 'i' ) {
    lastSensorType = SENSOR_INIT;
    RS_LOG( LA_SENSORS, "COACH INIT: " << sd );
  } else if ( sd[1] == 'h' ) {
    lastSensorType = SENSOR_AURAL_PLAYER;
    RS_LOG( LA_SENSORS, "COACH AURAL_PLAYER: " << sd );
  } else if ( sd[3] == 'e' ) {
    lastSensorType = SENSOR_VISUAL;
    RS_LOG( LA_SENSORS, "COACH SEE: " << sd );
    analyzeCoachSeeMessage(sd);
  } else {
    lastSensorType = SENSOR_COACH;
    RS_LOG( LA_SENSORS, "COACH UNKNOWN: " << sd );
  }
    
  theSensorInterface->setLastSensorType(lastSensorType);
}


void
Sensors::analyzeCoachSeeMessage(char* sightInfo)
{
  static ObjectType object_type;  

  static Float32U posX;
  static Float32U posY;
  static Float32U speedX;
  static Float32U speedY;
  static Float32U bodyDir;
  static Float32U neckDir;
  static UInt16U player_number;
  static char player_team_name[64];
  static Side player_side;
  static Team player_team;
  static MarkerName marker;

  RS_LOG( LA_SENSORS, sightInfo );
  
  if ( sightInfo[2] != 'e' )
    return;
  
  theSensorInterface->setCoachTime(getInt(&sightInfo));

  while (*sightInfo != ')') {

    posX.makeUnknown();
    posY.makeUnknown();
    speedX.makeUnknown();
    speedY.makeUnknown();
    bodyDir.makeUnknown();
    neckDir.makeUnknown();
    player_number.makeUnknown();
    player_team_name[0] = '\0';
    player_side = S_UNKNOWN;
    player_team = T_UNKNOWN;
    marker = UNKNOWN_MARKER;
 
    //std::cerr << sightInfo << std::endl; //"Before getWord...\n";
    sightInfo += 3;          // " ((" 
    //getWord(&sightInfo);            // " ((" 
    //std::cerr << sightInfo << std::endl;
    //std::cerr << "After getWord...\n";
    //std::cerr << ".";
    
    if ( *sightInfo=='g' ) {
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "goal " 
      if ( *sightInfo=='r' )
	marker = GOAL_R;
      else if ( *sightInfo=='l' )
	marker = GOAL_L;
      else
	RS_ERROR("goal ?");
    } else if ( *sightInfo=='f' ){
      object_type = OBJ_MARKER;
      sightInfo+=5;          // "flag " 
      if ( *sightInfo=='r' ){
	sightInfo+=2;
	if ( *sightInfo=='0' )
	  marker = FLAG_R0;
	else if ( *sightInfo=='b' ){
	  sightInfo+=1;
	  if ( *sightInfo==')' )
	    marker = FLAG_RB; 
	  else {
	    sightInfo+=1;
	    if ( *sightInfo=='1' )   marker = FLAG_RB10; else
	      if ( *sightInfo=='2' )   marker = FLAG_RB20; else
		if ( *sightInfo=='3' )   marker = FLAG_RB30; else
		  RS_ERROR("flag r b ?");
	  }  
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=1;
	    if ( *sightInfo==')' )     marker = FLAG_RT; 
	    else{
	      sightInfo+=1;
	      if ( *sightInfo=='1' )   marker = FLAG_RT10; else
		if ( *sightInfo=='2' )   marker = FLAG_RT20; else
		  if ( *sightInfo=='3' )   marker = FLAG_RT30; else
		    RS_ERROR("flag r t ?");
	    }  
	  } else
	    RS_ERROR("flag r ?");
      } else
	if ( *sightInfo=='l' ){
	  sightInfo+=2;
	  if ( *sightInfo=='0' )       marker = FLAG_L0;   else
	    if ( *sightInfo=='b' ){
	      sightInfo+=1;
	      if ( *sightInfo==')' )     marker = FLAG_LB; 
	      else{
		sightInfo+=1;
		if ( *sightInfo=='1' )   marker = FLAG_LB10; else
		  if ( *sightInfo=='2' )   marker = FLAG_LB20; else
		    if ( *sightInfo=='3' )   marker = FLAG_LB30; else
		      RS_ERROR("flag l b ?");
	      }  
	    } else
	      if ( *sightInfo=='t' ){
		sightInfo+=1;
		if ( *sightInfo==')' )     marker = FLAG_LT; 
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='1' )   marker = FLAG_LT10; else
		    if ( *sightInfo=='2' )   marker = FLAG_LT20; else
		      if ( *sightInfo=='3' )   marker = FLAG_LT30; else
			RS_ERROR("flag l t ?");
		}  
	      } else
		RS_ERROR("flag l ?");
	} else
	  if ( *sightInfo=='t' ){
	    sightInfo+=2;
	    if ( *sightInfo=='0' )       marker = FLAG_T0;   else
	      if ( *sightInfo=='l' ){
		sightInfo+=2;
		if ( *sightInfo=='1' )     marker = FLAG_TL10; else
		  if ( *sightInfo=='2' )     marker = FLAG_TL20; else
		    if ( *sightInfo=='3' )     marker = FLAG_TL30; else
		      if ( *sightInfo=='4' )     marker = FLAG_TL40; else
			if ( *sightInfo=='5' )     marker = FLAG_TL50; else
			  RS_ERROR("flag t l ?");
	      } else
		if ( *sightInfo=='r' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_TR10; else
		    if ( *sightInfo=='2' )     marker = FLAG_TR20; else
		      if ( *sightInfo=='3' )     marker = FLAG_TR30; else
			if ( *sightInfo=='4' )     marker = FLAG_TR40; else
			  if ( *sightInfo=='5' )     marker = FLAG_TR50; else
			    RS_ERROR("flag t r ?");
		} else
		  RS_ERROR("flag t ?");
	  } else
	    if ( *sightInfo=='b' ){
	      sightInfo+=2;
	      if ( *sightInfo=='0' )       marker = FLAG_B0;   else
		if ( *sightInfo=='l' ){
		  sightInfo+=2;
		  if ( *sightInfo=='1' )     marker = FLAG_BL10; else
		    if ( *sightInfo=='2' )     marker = FLAG_BL20; else
		      if ( *sightInfo=='3' )     marker = FLAG_BL30; else
			if ( *sightInfo=='4' )     marker = FLAG_BL40; else
			  if ( *sightInfo=='5' )     marker = FLAG_BL50; else
			    RS_ERROR("flag b l ?");
		} else
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='1' )     marker = FLAG_BR10; else
		      if ( *sightInfo=='2' )     marker = FLAG_BR20; else
			if ( *sightInfo=='3' )     marker = FLAG_BR30; else
			  if ( *sightInfo=='4' )     marker = FLAG_BR40; else
			    if ( *sightInfo=='5' )     marker = FLAG_BR50; else
			      RS_ERROR("flag b r ?");
		  } else
		    RS_ERROR("flag b ?");
	    } else
	      if ( *sightInfo=='c' ){
		sightInfo+=1;
		if ( *sightInfo==')' )       marker = FLAG_C;
		else{
		  sightInfo+=1;
		  if ( *sightInfo=='b' )     marker = FLAG_CB; else
		    if ( *sightInfo=='t' )     marker = FLAG_CT; else
		      RS_ERROR("flag c ?");
		}
	      } else
		if ( *sightInfo=='p' ){
		  sightInfo+=2;
		  if ( *sightInfo=='r' ){
		    sightInfo+=2;
		    if ( *sightInfo=='t')      marker = FLAG_PRT; else
		      if ( *sightInfo=='c')      marker = FLAG_PRC; else
			if ( *sightInfo=='b')      marker = FLAG_PRB; else	    
			  RS_ERROR("flag p r ?");
		  } else
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t')      marker = FLAG_PLT; else
			if ( *sightInfo=='c')      marker = FLAG_PLC; else
			  if ( *sightInfo=='b')      marker = FLAG_PLB; else	    
			    RS_ERROR("flag p l ?");
		    } else
		      RS_ERROR("flag p ?");
		} else
		  if ( *sightInfo=='g' ){
		    sightInfo+=2;
		    if ( *sightInfo=='l' ){
		      sightInfo+=2;
		      if ( *sightInfo=='t' )     marker = FLAG_GLT; else
			if ( *sightInfo=='b' )     marker = FLAG_GLB; else
			  RS_ERROR("flag g l ?");
		    } else
		      if ( *sightInfo=='r' ){
			sightInfo+=2;
			if ( *sightInfo=='t' )     marker = FLAG_GRT; else
			  if ( *sightInfo=='b' )     marker = FLAG_GRB; else
			    RS_ERROR("flag g r ?");
		      } else
			RS_ERROR("flag g ?");
		  } else
		    RS_ERROR("flag ?");
    } else
      if ( *sightInfo=='p' ){
	object_type = OBJ_PLAYER;
	sightInfo+=6;                          // "player" 
	if ( *sightInfo == ' ' ){              // there's a team  
	  sightInfo++;
	  // get the team name
	  int a = 0;
	  while ( *sightInfo != ' ' && *sightInfo != ')' && a < 63 )
	    player_team_name[a++]=*sightInfo++;
	  if ( a == 64 )
	    RS_WARNING("team name WAY to long > 63 characters!");
	  player_team_name[a] = '\0';
	    
	  if ( !strcmp(player_team_name, theSensorInterface->getOurTeamName().c_str()) ) {
	    player_side = theSensorInterface->getOurSide();
	    player_team = T_OUR_TEAM;
	  } else if ( theSensorInterface->getTheirTeamName() == "" ) {
	    player_side = theSensorInterface->getTheirSide();
	    player_team = T_THEIR_TEAM;
	  } else if ( !strcmp(player_team_name,
			      theSensorInterface->getTheirTeamName().c_str()) ) {
	    player_side = theSensorInterface->getTheirSide();
	    player_team = T_THEIR_TEAM;
	  }
	    
	  if ( *sightInfo== ' ' ){             // there's a number 
	    player_number = getInt(&sightInfo);
	  }
	}
      } else
	  if ( *sightInfo=='b' ) {
	    object_type = OBJ_BALL;
	  }
	  else {
	    RS_ERROR("unknown object in analyzeCoachMessage " + std::string(sightInfo));
	    return;
	  }
    
    advanceTo(')',&sightInfo);              // advance to end of object 

    /************************************/

    // All objects should have an absolute position
    posX = getFloat(&sightInfo);
    posY = getFloat(&sightInfo);

    // Players have absolute body direction and and absolute neck direction too
    if ( object_type == OBJ_PLAYER ) {
      bodyDir = getFloat(&sightInfo);
      neckDir = getFloat(&sightInfo);
    }

    // All moving objects have an absolute speed
    if ( object_type == OBJ_PLAYER ||
	 object_type == OBJ_BALL ) {
      speedX = getFloat(&sightInfo);
      speedY = getFloat(&sightInfo);
    }

    sightInfo++;                                // ")" 

    /************************************/

    //std::cerr << "Before switch...\n";
    
    switch (object_type){
    case OBJ_MARKER:
      theSensorInterface->coachSeeMarker(marker, posX, posY);
      break;
    case OBJ_BALL:
      theSensorInterface->coachSeeBall(posX, posY, speedX, speedY);
      break;
    case OBJ_PLAYER:
      theSensorInterface->coachSeePlayer(player_side, player_team, player_number,
					 posX, posY, speedX, speedY, bodyDir, neckDir);
      break;
    default:
      RS_ERROR("Unknown object in analyzeCoachMessage!");
    }
    //std::cerr << "sightInfo = " << sightInfo << std::endl;
  }
  //std::cerr << "sightInfo = " << sightInfo << std::endl;

  theSensorInterface->setLastSensorType(SENSOR_COACH);
}


/****************************************************************************/
/* These routines are to save time instead of using sscanf or atof, etc.    */
/* When passing **str_ptr, the pointer is advanced past the number          */
/* When passing  *str    , the pointer remains where it was before          */
/****************************************************************************/

// They are taken from CMUnited'98 created by Peter Stone et al 

Float32 Sensors::getFloat(char** str_ptr){

  Float32 d_frac;
  Float32 result;
  int m_flag;
  int d_flag;

  d_frac = 1.0;
  result = 0.0;
  m_flag = d_flag = 0;

  // Advance to the beginning of the number 
  while( !isdigit(**str_ptr) && **str_ptr!='-' && **str_ptr!='.')
    (*str_ptr)++;

  // Process the number bit by bit 
  while((**str_ptr!=')') && (**str_ptr!=' ') && (**str_ptr!='\n') && (**str_ptr!=',')){
    if (**str_ptr=='.')
      d_flag=1;
    else if (**str_ptr=='-')
      m_flag=1;
    else if (d_flag){
      d_frac *= 10.0;
      result+=(Float32)(**str_ptr-'0')/d_frac;
    }
    else
      result=result*10.0+(Float32)(**str_ptr-'0');
    (*str_ptr)++;
  }
  if (m_flag)
    result=-result;

  //printf("%.1f\n",result);

  return result;
}


// get the number, but don't advance pointer
Float32 Sensors::getFloat(char* str){
  char **str_ptr = &str;
  return getFloat(str_ptr);
}


Int32 Sensors::getInt(char** str_ptr)
{
  Int32 result;
  Int32 m_flag;

  result = 0;
  m_flag = 0;

  // Advance to the beginning of the number
  while( !isdigit(**str_ptr) && **str_ptr!='-')
    (*str_ptr)++;

  // Process the number bit by bit
  while((**str_ptr!=')') && (**str_ptr!=' ') &&
	(**str_ptr!='\n') && (**str_ptr!=',')) {
    if (**str_ptr=='-')
      m_flag=1;
    else
      result=result*10+(Int32)(**str_ptr-'0');
    (*str_ptr)++;
  }
  if (m_flag)
    result=-result;

  return result;
}


int Sensors::getInt(char* str)
{
  char **str_ptr = &str;
  return getInt(str_ptr);
}


void Sensors::getWord(char** str_ptr)
{
  while ( !isalpha(**str_ptr) ) (*str_ptr)++;
}


void Sensors::getNextWord(char** str_ptr)
{
  while ( isalpha(**str_ptr) ) (*str_ptr)++;
  getWord(str_ptr);
}


void Sensors::advanceTo(char c, char** str_ptr)
{
  while ( **str_ptr != c ) (*str_ptr)++;
}


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler or debugger does not understand the keyword inline then
 * include the inline definitions here, otherwise include them in the
 * declaration file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifndef RS_USE_INLINE
#  include "Sensors.icc"
#endif
