/* -*- Mode: C++ -*-
 * $Id: Sensors.h,v 1.10 2001/11/15 16:48:48 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: Sensors.h
 * Unitname: BasicSystem
 * $Revision: 1.10 $
 * Created by: Fredrik Heintz 1999-xx-xx
 * Last modified by $Author: frehe $ $Date: 2001/11/15 16:48:48 $
 *
 *
 * HISTORY
 *
 * $Log: Sensors.h,v $
 * Revision 1.10  2001/11/15 16:48:48  frehe
 * Added support for a coach.
 *
 * Revision 1.9  2001/09/17 21:54:00  frehe
 * Readded support for a debug coach.
 *
 * Revision 1.8  2001/09/17 21:30:09  andbo928
 * Implemented an ugly workaround to get ballspeeds working.
 *
 * Revision 1.7  2001/08/19 23:32:21  frehe
 * Updated 2001-08-19
 *
 * Revision 1.6  2001/08/17 07:15:13  frehe
 * Made changes to compile RoboSoc with RCCParser-0.10.
 *
 * Revision 1.5  2001/08/13 16:43:14  frehe
 * Made updates to compile with new autotools and RCCParser.
 *
 * Revision 1.4  2001/08/13 07:33:16  frehe
 * Updated from laptop 2001-08-13.
 *
 * Revision 1.3  2001/08/02 17:04:22  frehe
 * Updated 2001-08-02
 *
 * Revision 1.2  2000/09/04 08:46:45  frehe
 * Fixed a naming error in the sensors.
 *
 * Revision 1.1  2000/09/03 20:41:10  frehe
 * Imported the current version of RoboSoc (soon to be v2.5.0)
 *
 */

/**
 * @file Sensors.h
 * Include file for the class Sensors.
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.10 $
 * $Author: frehe $
 * $Date: 2001/11/15 16:48:48 $
 */


#ifndef SENSORS_H
#define SENSORS_H


#include <queue>

#include "rccparser.h"

#include "basic_system_types.h"
#include "ServerCommand.h"


RS_BEGIN_NAMESPACE


class Sensors;
class Controller;
class SensorInterface;


class SensorDataHandler
  : public RCC_Handler < RCC_Pointer < RCC_SenseBody > >,
    public RCC_Handler < RCC_Pointer < RCC_ServerParam > >,
    public RCC_Handler < RCC_Pointer < RCC_PlayerParam > >,
    public RCC_Handler < RCC_Pointer < RCC_PlayerType > >,
    public RCC_Handler < RCC_Pointer < RCC_Visual > >,
    public RCC_Handler < RCC_Pointer < RCC_RefAudio > >,
    public RCC_Handler < RCC_Pointer < RCC_SelfAudio > >,
    public RCC_Handler < RCC_Pointer < RCC_PlayerAudio > >,
    public RCC_Handler < RCC_Pointer < RCC_CoachAudio > >,
    public RCC_Handler < RCC_Pointer < RCC_Score > >,
    public RCC_Handler < RCC_Pointer < RCC_InitMsg > >,
    public RCC_Handler < RCC_Pointer < RCC_Error > >,
    public RCC_Handler < RCC_Pointer < RCC_Substitution > > 
{
public:
  SensorDataHandler(Sensors* _sensors);
  
  void handle ( const RCC_Pointer < RCC_SenseBody >& sb_p );
  void handle ( const RCC_Pointer < RCC_ServerParam >& sp_p );
  void handle ( const RCC_Pointer < RCC_PlayerParam >& pp_p );
  void handle ( const RCC_Pointer < RCC_PlayerType >& pt_p );
  void handle ( const RCC_Pointer < RCC_Visual >& pt_p );
  void handle ( const RCC_Pointer < RCC_RefAudio >& ra_p );
  void handle ( const RCC_Pointer < RCC_SelfAudio >& sa_p );
  void handle ( const RCC_Pointer < RCC_PlayerAudio >& pa_p );
  void handle ( const RCC_Pointer < RCC_CoachAudio >& ca_p );
  void handle ( const RCC_Pointer < RCC_Score >& score_p );
  void handle ( const RCC_Pointer < RCC_InitMsg >& init_p );
  void handle ( const RCC_Pointer < RCC_Error >& error_p );
  void handle ( const RCC_Pointer < RCC_Substitution >& sub_p );
  
protected:
  PlayMode getPlayMode(const RCC_StadPlayMode& play_mode) const;
  ViewQuality getViewQuality(const RCC_ViewMode::quality_t& rcc_view_quality) const;
  ViewWidth getViewWidth(const RCC_ViewMode::width_t& rcc_view_width) const;
  CoordPoint<Float32U> getVectorUnknown(const RCC_Vector2d& vec) const;
  CoordPoint<Float32U> getPointUnknown(const RCC_Vector2d& vec) const;
  SideLineName getSideLineName(const RCC_VisualObject::location_t& loc) const;
  MarkerName getGoalName(const int l) const;
  MarkerName getFlagName(const int l) const;
  void updateTime(int time);
  void setLastSensorType(SensorType sensor_type) throw();
  
private:
  Sensors* sensors;
};


/**
 * @class Sensors Sensors.h
 * This class is responsible for parsing the server messages
 * and storing them in the sensor interface for further use
 * by the agent.
 *
 * <!-- @invariant The class invariant is... -->
 *
 * @par Unit:
 *      BasicSystem
 * $Revision: 1.10 $
 * $Author: frehe $
 * $Date: 2001/11/15 16:48:48 $
 */
class Sensors
{
public:
  Sensors(SensorInterface*);

  void init(Controller*);

  /** Analyze messages from the server to the player. */
  void analyze(char*);
  /** Analyze coach messages. */
  void analyzeCoachMessage(char*);
  void analyzeCoachSeeMessage(char*);
  /** Analyze feedback from the actuators. */
  void analyze(ServerCommand*);  
  // The feedback should only be updated at the same time as the agents
  // are updated in the server. I.e. Turn neck, change view a.s.o. are
  // updated when they arrive while dash, kick and turn are only updated
  // at the end of the cycle.
  // But isn't that something for the views to take care of?
  // If the sensor interface stores the commands, then it's up to the
  // views when to do the actual update.
  // Major problem that only one command can be stored in the sensor interface.
  // Or maybe not, since only one command is executed last in the cycle
  // that's the only one that needs to be saved. But it could be a problem.

  SensorType getLastSensorType() const throw();
  void setLastSensorType(SensorType sensor_type) throw();
  UInt16 getCurrentTime() const throw();
  void setCurrentTime(UInt16 new_time) throw();

protected:
  friend class RS_NAMESPACE::SensorDataHandler;
  
  Controller* theController;
  SensorInterface* theSensorInterface;
  RCC_Parser theParser;

  SensorDataHandler sensorDataHandler;

  SensorType lastSensorType;
  UInt16 currentTime;

  std::queue<CommandType> commandQueue;
  
  enum ObjectType {
    OBJ_LINE,
    OBJ_BALL,
    OBJ_MARKER,
    OBJ_MARKER_BEHIND,
    OBJ_PLAYER
  };
  
  Float32 getFloat(char** str_ptr);
  Float32 getFloat(char* str);
  Int32 getInt(char* str);
  Int32 getInt(char** str_ptr);
  void getWord(char** str_ptr);
  void getNextWord(char** str_ptr);
  void advanceTo(char c, char** str_ptr);
  
  // Hide the copy-constructor and the assignment operator.
  Sensors(const Sensors&);
  Sensors& operator= (const Sensors&);
};


RS_END_NAMESPACE


/* CONDITIONAL INCLUSION OF INLINE DEFINITIONS
 *
 * If the compiler and debugger understands the keyword inline then include
 * the inline definitions here, otherwise include them in the definition file.
 *
 * This is controlled by a flag called RS_USE_INLINE, which is usually
 * defined. It can be set by the user by giving the compiler an argument
 * usually -DRS_USE_INLINE (to define it) or -URS_USE_INLINE (to undefine it).
 */

#ifdef RS_USE_INLINE
#  include "Sensors.icc"
#endif


#endif // !SENSORS_H
