/* -*- Mode: C++ -*-
 * $Id: FFClearBall.cc,v 1.3 2001/09/29 18:55:20 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *0
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FFClearBall.cc
 * Unitname: Framework/skill
 * $Revision: 1.3 $
 * Created by:Chris & Govert
 * Last modified by $Author: frehe $ $Date: 2001/09/29 18:55:20 $
 *
 *
 * HISTORY
 */

/**
 * @file FFClearBall.cc
 * Implementation file for the skill class FFClearBall.
' *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/09/29 18:55:20 $
 */


#include "AngleThingy.h"
#include "FFClearBall.h"
#include "ViewManager.h"
#include <list>
#include <algorithm>

RS_BEGIN_NAMESPACE

/**
 * This class describes the clearball skill. It's called in an emergency (probably 
 * in front of the own goal) and it results in kicking the ball to an open area on 
 * the field (preferably towards the enemy side of the field)
 *
 * Preconditions
 *  The player needs to be in kicking distance to the ball
 *  It is impossible to give a good pass to a friendly player
 *  It is to dangerous to try to dribble (e.g. if the enemy players get the ball they will score)
 * Postconditions
 *  The ball is kicked to an empty area on the field if an appropriate area exists.
 *  if no good area exists within the view of the player he tries to kick it in the direction 
 *  of the enemy goal
 *  the ball is NEVER kicked in the direction of the own side of the field (because this could 
 *  cause a more dangerous situation)
 *  The ball is only kicked outside of the field if there is no better solution
 */
FFClearBall::FFClearBall()
  : Skill(),
    worldView(NULL)
{
  getView(worldView, WORLD_STATE_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
} 


FFClearBall::~FFClearBall()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
}

//------------------------------------------------------------

// --- METHODS ---

Plan FFClearBall::generatePlan()
{
  // Build anglethingy.
  AngleThingy angles;

  // Initalize anglethingy.
  const AgentState& agent
    = worldView->getWorldState().getAgent();
  BallState ball
    = worldView->getWorldState().getBall();

  // Assume max kick.
  double kpr = agent.getKickPowerRate(ball);
  double v0 = kpr * 100.0;
  angles.setInitialSpeed(v0);

#define WIDTH 65
#define LENGTH 102
#define KICK_RANGE 60
#define MAX_ANG 45

  // Find the valid range.
  TLine leftLine(TVector(-LENGTH/2, -WIDTH/2),
		 TVector(LENGTH/2,  -WIDTH/2));
  TLine rightLine(TVector(LENGTH/2,  WIDTH/2),
		  TVector(-LENGTH/2, WIDTH/2));
  TVector leftLimit = TVector::TPolarVector(KICK_RANGE, -MAX_ANG);
  TVector rightLimit = TVector::TPolarVector(KICK_RANGE, MAX_ANG);

  TVector dist = leftLine.getDistance(ball.getPosition());
  // Trim left side.
  if(dist.getLength() < KICK_RANGE * std::sin(MAX_ANG*M_PI/180.0))
  {
//      cout << "Trim left: " << dist << endl;
    float otherLen = sqrt(KICK_RANGE*KICK_RANGE - dist*dist);
    leftLimit = -dist + TVector(1,0) * otherLen;
  }

//    // Trim right side.
  dist = rightLine.getDistance(ball.getPosition());
  if(dist.getLength() < KICK_RANGE * std::sin(MAX_ANG*M_PI/180.0))
  {
//      cout << "Trim right: " << dist << endl;
    float otherLen = sqrt(KICK_RANGE*KICK_RANGE - dist*dist);
    rightLimit = -dist + TVector(1,0) * otherLen;
  }

  AngleThingy::Span initSpan;

  if(ball.getPosition().getX() < 10)
  {
    initSpan.start = leftLimit.getAngle();
    initSpan.end = rightLimit.getAngle();
  }
  else
  {
    // Find the initial span, the span between the two goalposts in this case.
    TVector leftLimit = 
      TVector(parameterView->SP_pitch_length/2,
	      -0.95*parameterView->SP_goal_width/2) - ball.getPosition();
    TVector rightLimit = 
      TVector(parameterView->SP_pitch_length/2,
	      0.95*parameterView->SP_goal_width/2) - ball.getPosition();
    
    initSpan.start = leftLimit.getAngle();
    initSpan.end = rightLimit.getAngle();
  }
  angles.addSpan(initSpan);

  // Time for heavy black magic.
  // Split the angles
  float bestDir = 0.0;

  const int maxSteps = 40;
  const int actionPenalty = 3;
  int time;
  for(time = 0; time < maxSteps; time++)
  {
    angles.step();

    // Find the best "safe" direction.
    if(angles.getSpans().size() == 0)
      break;

    AngleThingy::Span bestSpan = angles.getMaxSpan();
    bestDir = (bestSpan.start + bestSpan.end) / 2.0;
  }
//    cout << "Bailout, time: " << time 
//         << ", best dir: " << bestDir << endl;

//    TAngle relDir = bestDir 
//      - worldView->getWorldState().getAgent().getDirection();
//    TVector kickVector = TVector::TPolarVector(100.0, relDir);
//    WorldState ws = worldView->getWorldState();
//    ws.doKick(kickVector);
//    Plan p = Plan(ws.getStep());

  TVector kickVector = TVector::TPolarVector(3.0, bestDir);
  WorldState ws = worldView->getWorldState();
  ws.doKickToSpeed(kickVector);
  Plan p = Plan(ws.getStep());
  return p;
}

FuzzyBool FFClearBall::persistent()
{
  return FuzzyBool(0.0);
}

FuzzyBool FFClearBall::applicable()
{
  const WorldState& ws = worldView->getWorldState();
  if(ws.getAgent().isBallKickable(ws.getBall()))
    return FuzzyBool(1.0);
  return FuzzyBool(0.0);
}

FuzzyBool FFClearBall::succeed()
{
  return FuzzyBool(1.0);
}

FuzzyBool FFClearBall::failed()
{
  return FuzzyBool(0.0);
}



RS_END_NAMESPACE
