/* -*- Mode: C++ -*-
 * $Id: GoaliePositioning.cc,v 1.12 2001/10/22 16:32:51 aip11 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GoaliePositioning.cc
 * Unitname: Framework/skill
 * $Revision: 1.12 $
 * Created by: Fredrik Heintz 2000-09-07
 * Last modified by $Author: aip11 $ $Date: 2001/10/22 16:32:51 $
 *
 *
 * HISTORY
 *
 * $Log: GoaliePositioning.cc,v $
 * Revision 1.12  2001/10/22 16:32:51  aip11
 * Added backward moving to keep ball in view or for quicker movement, rush out more aggresivly,changed intercept extrapolation point
 *
 * Revision 1.11  2001/10/11 11:05:04  aip11
 * Changed some Plan constructors to use heap instead of stack
 *
 * Revision 1.10  2001/10/10 18:33:43  aip11
 * Added some comment and tweaked a few constants
 *
 * Revision 1.9  2001/09/29 18:55:20  frehe
 * Removed TRUE, FALSE, and UNKNOWN.
 *
 * Revision 1.8  2001/09/26 11:41:25  frehe
 * Update after move of TRUE, FALSE and UNKNOWN to inside FuzzyBool.
 *
 * Revision 1.7  2001/09/26 09:59:52  aip11
 * Removed some couts
 *
 * Revision 1.6  2001/09/26 09:03:52  aip11
 * Fixed confidence bug
 *
 * Revision 1.4  2001/09/20 12:59:48  aip11
 * Somewhat working version
 *
 * Revision 1.3  2001/09/18 16:28:29  aip11
 * *** empty log message ***
 *
 * Revision 1.2  2001/09/06 13:12:53  andbo928
 * Added some methods to the class. (SkillStub was bad)
 *
 * Revision 1.1  2001/09/05 14:41:03  stejo127
 * Added the GoaliePositioning skill.
 *
 * Revision 1.3  2000/09/16 18:39:39  frehe
 * Fixed the log messages.
 *
 * Revision 1.2  2000/09/16 18:36:48  frehe
 * Made minor fixes the skill stubs.
 *
 * Revision 1.1  2000/09/07 16:10:34  frehe
 * Added stubs for the skills.
 *
 */

/**
 * @file GoaliePositioning.cc
 * Implementation file for the skill class GoaliePositioning.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.12 $
7 * $Author: aip11 $
 * $Date: 2001/10/22 16:32:51 $
 */


#include "GoaliePositioning.h"
#include "ParameterView.h"
#include "MarkerView.h"
#include "AgentView.h"
//#include "BallView.h"
#include "view_constants.h"
#include "MoveToPosition.h"
#include "WorldStateView.h"

RS_BEGIN_NAMESPACE

// --- CONSTRUCTORS ---

GoaliePositioning::GoaliePositioning()
  : Skill(),
    parameterView(NULL),
    ballView(NULL),
    agentView(NULL),
    markerView(NULL)
{
  getView(parameterView, PARAMETER_VIEW_ID);  
  // getView(ballView, BALL_VIEW_ID);
  getView(markerView,MARKER_VIEW_ID, parameterView);
  //getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(worldStateView,WORLD_STATE_VIEW_ID);
}


// --- DESTRUCTOR ---

GoaliePositioning::~GoaliePositioning()
{
  //viewManager->releaseView(BALL_VIEW_ID);
  //viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
}


// --- METHODS ---   

Plan GoaliePositioning::generatePlan()
{
  WorldState worldState=worldStateView->getWorldState();
  BallState ball=worldState.getBall();
  AgentState agent=worldState.getAgent();
  TVector ballPos=ball.getPosition();
  TVector ballSpeed=ball.getSpeed();
  TVector goalCenter=TVector(markerView->getOurGoal(NOW).getPosition());
  static int lastCatch=0;
  lastCatch--;
  //Check if ball is catchable, if so catch it
  if((ballPos.getC()>0.8) && (ballPos-agent.getPosition()).getLength()<parameterView->SP_catchable_area_l*0.97 && lastCatch<0){
    CatchCommand* cc=new CatchCommand((ballPos-agent.getPosition()).getAngle()-agent.getDirection());
    Plan plan(cc);
    delete cc;
    lastCatch=5;
    return plan;
  }

  if(agent.isBallKickable(ball) && lastCatch>=0){
    float ypos=30;
    if(agent.getPosition().getY()<0)
      ypos=-30;
    agent.doKickToPosition(TVector(0,ypos),ball);  
    std::cerr << "Goalie Panic! \n";
    Step s=agent.getStep();
    return Plan(s);
  }
  //Check how close the opposition is to the ball
  float closeOpDist=1000;
  std::vector<PlayerState> opponents= worldState.getOpponents ();
  std::vector<PlayerState>::iterator psi;
  for(psi=opponents.begin();psi!=opponents.end();psi++){
    TVector relPos=ballPos-psi->getPosition();
    TVector relSpeed=ballSpeed-psi->getSpeed();
    if(relSpeed*relPos>0){
      relPos+=relSpeed*3;
    }
    float opDist=relPos.getLength();
    if(psi->getC()>0.7 && opDist<closeOpDist){
      closeOpDist=opDist;
    }
  }
  //Predict where to intercept ball
  TVector predBallPos;
  TVector apos=agent.getPosition();
  float predict=0;
  if(ballSpeed*(apos-ballPos)>0){
    TVector unitBallSpeed=ballSpeed.normalized();
    predict=(ballPos-apos+unitBallSpeed*(ballPos-apos).getLength()).getLength()*1.4;
    if(predict*ballSpeed.getLength()>(ballPos-apos).getLength())
      predict=(ballPos-apos).getLength()/ballSpeed.getLength();
  } else {
    predict=(ballPos-apos).getLength();
  }
  if(predict>closeOpDist*3)
    predict=closeOpDist*3;
  predBallPos=ballPos+ballSpeed*predict;
  if(predBallPos.getX()<-52)
    predBallPos.setX(-52);

  //do a second ball prediction to see if we want to rush out
  BallState predBall=ball;
  TVector predBallPos2=ball.getPosition();
  int maxPred=15;
  if(maxPred>closeOpDist*3)
    maxPred=(int)(closeOpDist*3);
  for(int a=0;a<maxPred;a++){
    predBall.step();
    if((predBall.getPosition()-goalCenter).getLength()<15){
      predBallPos2=predBall.getPosition();
      break;
    }
  }

  //The positioning is done here
  if(worldState.getBall().getPosition().getC()>0.6f){
    //Calculate the vector between the own goal and the ball
    TVector ballGoalVec = predBallPos-goalCenter;
    float ballDistance=ballGoalVec.getLength();
    //place the goalie on a 1/0.14-radius-circle from the center of the goal
    ballGoalVec/=ballDistance*0.14;
    TVector pos;
    float maxDist=1;
    float ballDistance2=(predBallPos2-goalCenter).getLength();

    if(ballDistance<15 || (predBallPos-agent.getPosition()).getLength()<4){
      pos=predBallPos;
      maxDist=0.3;
    } else if(ballDistance2<15){
      pos=predBallPos2;
      maxDist=0.3;
    } else {      //Place the goalie on circle
      pos=TVector(-54,0)+ballGoalVec;
    }
    //use more dashpower if ball is close
    float dashPower=140-ballDistance*2;
    if(dashPower>100)
      dashPower=100;
    else if (dashPower<5)
      dashPower=5;
    
    return MoveToPosition(pos,dashPower,maxDist);
  } else {
    //Turn around and try to find ball
    TurnCommand* tc=new TurnCommand(-45);
    Plan plan(tc);
    delete tc;
    return plan;
  }

}


FuzzyBool GoaliePositioning::persistent()
{
  // Never persistent.
  return FuzzyBool(0.0);
}

FuzzyBool GoaliePositioning::applicable()
{
  // Applicable if goalie
  return agentView->isGoalie();
}

FuzzyBool GoaliePositioning::succeed()
{
  // Always succedes.
  return FuzzyBool(1.0);
}

FuzzyBool GoaliePositioning::failed()
{
  // Never fails.
  return FuzzyBool(0.0);
}

///Our move to position-function, uses the FuzzyFoo-functions
Plan GoaliePositioning::MoveToPosition(const TVector& goal,const float power,float maxDist)
{
  WorldState worldState=worldStateView->getWorldState();
  AgentState agent=worldState.getAgent();
  BallState ball=worldState.getBall();

  TVector ballPos=ball.getPosition();
  TVector agentPos=agent.getPosition();
  TVector playerToGoal=goal-agentPos;
  TVector playerToBall=ballPos-agentPos;

  bool seeBack=true,seeFront=true;
  float relDir=playerToBall.normalized()*playerToGoal.normalized();
  if(relDir>0.6)
    seeBack=false;
  else if (relDir<-0.6)
    seeFront=false;

  float ang=playerToGoal.getAngle()-agent.getDirection();
  while(ang>180)
    ang-=360;
  while(ang<-180)
    ang+=360;
  
  float revAng=ang+180;
  while(revAng>180)
    revAng-=360;
  
  float powerDir=1;
  if(!seeFront || (seeBack && abs(revAng)<20)){
    powerDir=-1;
    ang=revAng;
  }

  if(playerToGoal.getLength()>maxDist){
    if ( abs(ang) > 20 ){
      agent.doActualTurn(ang);
    } else {
      if(playerToGoal.getLength()<agent.getSpeed().getLength()*(maxDist*3+1))
	agent.doDashToSpeed(0);
      else 
	agent.doDash(power*powerDir);
    }
  } else {
    agent.doTurnTowards(worldState.getBall().getPosition());
  }
  agent.doTurnNeckTowards(worldState.getBall().getPosition());
  Step s=agent.getStep();
  return Plan(s);
}

RS_END_NAMESPACE



