/* -*- Mode: C++ -*-
 * $Id: GotoPassablePosition.h,v 1.9 2001/10/10 14:50:17 aip12 Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: GotoPassablePosition.h
 * Unitname: Framework/skill
 * $Revision: 1.9 $
 * Created by: Peter Andersson and Jonas Carlsson
 * Last modified by $Author: aip12 $ $Date: 2001/10/10 14:50:17 $
 *
 *
 * HISTORY
 * $Log: GotoPassablePosition.h,v $
 * Revision 1.9  2001/10/10 14:50:17  aip12
 * SIFindBall added. /Peter
 *
 * Revision 1.8  2001/10/09 13:05:41  aip12
 * Added so that you can chose whether to be or not to be (that is the question) in the line of sight of the player with the ball.
 *
 * Revision 1.7  2001/10/08 19:13:19  aip12
 * Updated applicable. Now it checks if some teammate has the ball too.
 *
 * Revision 1.5  2001/10/02 17:34:11  aip12
 * Fixat lite funktionalitet, fungerar hyffsat i vissa situationer.
 *
 * Revision 1.4  2001/09/30 16:31:31  aip12
 * Updated skill, nicer code, still lots of debug info, compiles after changes with FuzzyBool
 *
 * Revision 1.3  2001/09/24 16:51:30  aip12
 * Generate plan is "complete" (NOT TESTED!!!). You can compile it.
 *
 * Revision 1.2  2001/09/18 15:21:10  aip12
 * Skill implemented but not tested.
 *
 * Revision 1.1  2001/09/02 10:43:16  petan117
 * Added skill skeleton for GotoPassablePosition /Peter o Jonas
 *
 *
 */

/**
 * @file GotoPassablePosition.h
 * Include file for the skill class GotoPassablePosition.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.9 $
 * $Author: aip12 $
 * $Date: 2001/10/10 14:50:17 $
 */
#ifndef GOTOPASSABLEPOSITION_H
#define GOTOPASSABLEPOSITION_H

#include "Skill.h"
#include "TVector.h"

#include <cmath>

#include "framework_types.h"
#include "SkillWithDirection.h"
#include "view_constants.h"
#include "WorldState.h"

RS_BEGIN_NAMESPACE
class WorldStateView;

/**
 * @class GotoPassablePosition GotoPassablePosition.h
 *  This is a skill to place a player without the
 *  ball in a position where it is the least chance
 *  for an opponent to intercept a pass from another
 *  player.
 * @pre The preconditions for the skill are:
 *  - The player doesn't have the ball
 *  - The player sees the ball
 *  - A teammate has the ball
 * @post The intended postconditions for the skill are:
 *  - The player moves to a more favourable position for a pass.
 *  - The player is in the line of site of the player with the ball (if setPlayerView(true) is called)
 * @todo Not go into offside position
 *   
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.9 $
 * $Author: aip12 $
 * $Date: 2001/10/10 14:50:17 $ */
class GotoPassablePosition : public Skill
{
public:
  /**
   * Default constructor. Here you can initialize the skill.
   */
  GotoPassablePosition();

  /**
   * Constructor for getting in a good position towards a given point instead of the ball position. 
   * Can be used to make passtrees.
   */
  GotoPassablePosition(TVector point);

  /**
   * Destructor should clean up after the skill.
   */
  virtual ~GotoPassablePosition();

  /**
   * Procedure to set if the agent must be in line of site of the teammate who has the ball.
   */
  void setPlayerView(bool InPlayerView);

  /**
   * This method is used to generate a plan for how the goal of
   * the skill should be obtained by the agent.
   */
  virtual Plan generatePlan();

  /**
   * This method should return a FuzzyBool telling the caller
   * if the skill wants to be called again.
   */
  virtual FuzzyBool persistent();

  /**
   * This method should implement the precondition specified in
   * the description of the skill.
   */
  virtual FuzzyBool applicable();

  
  /**
   * Return @c true (close to 1) if the skill has succeeded,
   * i.e. its post conditions are satisfied.
   */
  virtual FuzzyBool succeed();

  /**
   * Return @c true (close to 1) if the skill has failed,
   * i.e. something went wrong.
   */
  virtual FuzzyBool failed();

private:
  WorldStateView* worldStateView;
  bool inPlayerView;
  TVector gotoPassablePoint;

  Float dotProduct(TVector x, TVector y);
  double abs(double a);
  double getCoefSum(WorldState &worldState, TVector agentPosition,bool inPlayerView);
  double getVectorsLength(TVector v1, TVector v2, TVector v3);
  Step doStep(TVector posVector, AgentState &agentState);
  PlayerState *hasTeammateBall (WorldState &worldState);

};

RS_END_NAMESPACE

#endif // !GOTOPASSABLEPOSITION_H




