
/*
 * $Id: WorldState.cc,v 1.1 2001/09/18 20:41:50 andbo928 Exp $
 * 
 * HISTORY 
 *
 * $Log: WorldState.cc,v $
 * Revision 1.1  2001/09/18 20:41:50  andbo928
 * Imported some sources from FuzzyFoo.
 *
 * Revision 1.48  2001/07/29 20:53:57  sigge
 * Added WorldState::getInterceptors and appropriate methods in StatisticsView to accelerate interception calculations.
 *
 * Revision 1.47  2001/07/28 14:07:56  andbo928
 * Removed spam.
 *
 * Revision 1.46  2001/07/28 13:39:13  andbo928
 * Removed a ton of shadowplayers.
 *
 * Revision 1.45  2001/07/28 09:22:46  mikbr212
 * Cosmetic changes
 *
 * Revision 1.44  2001/07/28 06:50:53  mikbr212
 * Removed warnings
 *
 * Revision 1.43  2001/07/27 13:20:31  sigge
 * Do not model opponents (this crashes).
 *
 * Revision 1.42  2001/07/27 13:02:32  mikbr212
 * Added stepR
 *
 * Revision 1.41  2001/07/27 10:32:02  mikbr212
 * Turned modelling on
 *
 * Revision 1.40  2001/07/26 18:59:44  mikbr212
 * Added modelling, but it causes a segmentation fault - therefore disabled
 *
 * Revision 1.39  2001/07/26 13:47:53  sigge
 * Added offside helper functions.
 *
 * Revision 1.38  2001/07/12 12:45:35  andbo928
 * Now uses a more generous match function to reduce shadowplayers.
 *
 * Revision 1.37  2001/07/10 17:29:00  andbo928
 * Reimplemented the mergePlayers method.
 *
 * Revision 1.36  2001/07/06 20:48:47  mikbr212
 * Added getTeam
 *
 * Revision 1.35  2001/07/06 14:27:46  andbo928
 * Fixed field-mirroring.
 *
 * Revision 1.34  2001/07/06 13:02:32  sigge
 * Removed IntentionDecision from various places.
 *
 * Revision 1.33  2001/07/06 10:00:39  mikbr212
 * Merged updates and resolved conflicts (a lot)
 *
 * Revision 1.32  2001/07/05 17:12:23  andbo928
 * Added transformation code.
 *
 * Revision 1.31  2001/07/03 12:46:26  sigge
 * Linking problem is fixed.
 *
 * Revision 1.30  2001/07/02 14:44:10  andbo928
 * Nothing much.
 *
 * Revision 1.29  2001/07/01 12:39:52  andbo928
 * Added new limitations imposed by the new server in simulation. Added doChangeViewFoo methods.
 *
 * Revision 1.28  2001/06/19 21:41:54  mikbr212
 * Added comments
 *
 * Revision 1.27  2001/06/19 14:03:35  mikbr212
 * Added a shortcut: setIntention(Intention*) to agent.setIntention(Intention*)
 *
 * Revision 1.26  2001/06/05 20:33:53  mikbr212
 * Fixed doxygen-stubs for all classes
 *
 */

/**
 * @file WorldState.cc
 * Implementation file for the class WorldState.
 *
 * @par Unit:
 *      Framework/view
 * $Revision: 1.1 $
 * $Author: andbo928 $
 * $Date: 2001/09/18 20:41:50 $
 */

#include "WorldState.h"

#include <stdio.h>
#include <iostream>
#include "Step.h"

#include <list>

RS_BEGIN_NAMESPACE

/** 
 * Constructor. 
 * Constructs an empty WorldState
 */
WorldState::WorldState() :
  State(),
  parameterView(NULL),
  gameTime(0),
  team(T_OUR_TEAM)
  
{
  getView(parameterView, PARAMETER_VIEW_ID);

  ball.setPosition(TVector(0,0,0));
  agent.setPosition(TVector(0,0,0));

  opponents.clear();
  teammates.clear();
}

/**
 */
WorldState::~WorldState()
{
  ViewManager::instance()->releaseView(PARAMETER_VIEW_ID);  
}

/**
 */
const std::vector<PlayerState>& WorldState::getOpponents() const
{
  return opponents;
}

/**
 */
void WorldState::setOpponents(const std::vector<PlayerState>& vec)
{
  opponents = vec;
}

/**
 */
const std::vector<PlayerState>& WorldState::getTeammates() const
{
  return teammates;
}

/**
 */
void WorldState::setTeammates(const std::vector<PlayerState>& vec)
{
  teammates = vec;
}

/**
 */
const BallState& WorldState::getBall() const
{
  return ball;
}

/**
 */
void WorldState::setBall(const BallState& other)
{
  ball = other;
}

/**
 */
const AgentState& WorldState::getAgent() const
{
  return agent;
}

/**
 */
void WorldState::setAgent(const AgentState& other)
{
  agent = other;
}

/**
 */
int WorldState::getGameTime() const
{
  return gameTime;
}

/** Returns the team beloning of the agent and the teammates.
 * Unless a changeAgent (like in modelling) has been performed
 * the team should be T_OUR_TEAM.
 */
Team WorldState::getTeam() const {
  return team;
}

/**
 */
void WorldState::setGameTime(int gt)
{
  gameTime = gt;
  ball.setGameTime(gt);
  agent.setGameTime(gt);
  for(unsigned int i = 0; i < teammates.size(); i++)
    teammates[i].setGameTime(gt);
  for(unsigned int i = 0; i < opponents.size(); i++)
    opponents[i].setGameTime(gt);
}

/**
 */
void WorldState::changeFrame(PlayerState& plr,
			     const TVector& frameDelta,
			     const float frameDirChange,
			     const TVector& frameOrigin)
{
  // Update position.
  TVector plrPos = plr.getPosition();
  plrPos = plrPos + frameDelta - frameOrigin;
  plrPos.rotate(frameDirChange);
  plr.setPosition(plrPos + frameOrigin);

  // Update speed.
  plr.setSpeed(plr.getSpeed().rotate(frameDirChange));

  // Update direction.
  plr.setDirection(plr.getDirection() + frameDirChange);
}

/**
 */
void WorldState::changeFrame(BallState& ball,
			     const TVector& frameDelta,
			     const float frameDirChange,
			     const TVector& frameOrigin)
{
  // Update position.
  TVector ballPos = ball.getPosition();
  ballPos = ballPos + frameDelta - frameOrigin;
  ballPos.rotate(frameDirChange);
  ball.setPosition(ballPos + frameOrigin);

  // Update speed.
  ball.setSpeed(ball.getSpeed().rotate(frameDirChange));
}

/**
 */
void WorldState::framingCompensate(const WorldState& other)
{
  // Assumes that this worldstate is synced to the other.
  TVector frameDelta = 
    other.agent.getPosition() - agent.getPosition();
  float frameDirChange = 
    other.agent.getDirection() - agent.getDirection();
  TVector frameOrigin = other.agent.getPosition();

  changeFrame(agent, frameDelta, frameDirChange, frameOrigin);
  changeFrame(ball, frameDelta, frameDirChange, frameOrigin);

  for(unsigned int i = 0; i < opponents.size(); i++)
    changeFrame(opponents[i], frameDelta, frameDirChange, frameOrigin);
  for(unsigned int i = 0; i < teammates.size(); i++)
    changeFrame(teammates[i], frameDelta, frameDirChange, frameOrigin);

}

/** Checks if the two players matches each other, ie are almost
 * the same. 
 * @return a value between 0 and 1, where 1 is a perfect match and 
 * 0 is no match.
 */
double WorldState::matchPlayer(const PlayerState& plr1,
			       const PlayerState& plr2)
{
  // Check if any unum is known.
  if(plr1.getUniformNumber() != -1 || plr2.getUniformNumber() != -1) {

    if(plr1.getUniformNumber() == plr2.getUniformNumber())
    {
      return 1.0; // "Perfect" match.
    }
    
    if(plr1.getUniformNumber() != -1 && plr2.getUniformNumber() != -1)
    {
      return 0.0; // "Perfect" anti-match.
    }
  }
  
  TVector delta = plr1.getPosition() - plr2.getPosition();
  double d = 0.9 * exp( -0.0003 * (delta*delta));
  return d;
}

void printList(const std::list<PlayerState>& lst)
{
  std::list<PlayerState>::const_iterator it;
  cout << "{";
  for(it = lst.begin(); it != lst.end(); it++)
    cout << it->getUniformNumber() << ", ";
  cout << "}" << endl;
}

void WorldState::fixConsitency(std::list<PlayerState>& players)
{

  // Threshold at which two players are considerd
  // to be the same player if at most one unum is known.
  const double sameThr = 2.0;  

  // First check if we find any duplicate unums.
  // Due to stupid roboserver we assume that
  // players that are later in the list are more accurate.

//    if(agent.getUniformNumber() == 1)
//    {
//      cout << "Pre same-num-fix." << endl;
//      printList(players);
//    }

  while(players.size() > 10)
  {
    players.pop_front();
  }

  std::list<PlayerState>::iterator itP;
  itP = players.begin();
  while(itP != players.end())
  {
    int unum1 = itP->getUniformNumber();
    if(unum1 != -1)
    {
      std::list<PlayerState>::iterator itP2 = itP;
      itP2++;
      while(itP2 != players.end())
      {
	if(unum1 == itP2->getUniformNumber())
	{
	  // Later is better, se above.
	  *itP = *itP2;
	  itP2 = players.erase(itP2);
	}
	else
	{
	  itP2++;
	}
      }
    }
    itP++;
  }

//    if(agent.getUniformNumber() == 1)
//    {
//      cout << "Post same-num-fix / Pre same-pos-fx." << endl;
//      printList(players);
//    }

  // Now we remove players that are considerd to be the same.
  itP = players.begin();
  while(itP != players.end())
  {
    int unum1 = itP->getUniformNumber();
    std::list<PlayerState>::iterator itP2 = itP;
    itP2++;
    while(itP2 != players.end())
    {
      TVector delta = itP2->getPosition() - itP->getPosition();
      int unum2 = itP2->getUniformNumber();
      if( unum1 == -1 || unum2 == -1 )
      {
	if(delta*delta < sameThr*sameThr)
	{
	  // They're the same.
	  *itP = *itP2;
	  itP2 = players.erase(itP2);

	  if(unum1 != -1)
	    itP->setUniformNumber(unum1);
	  else
	    itP->setUniformNumber(unum2);
	} 
	else
	{
	  itP2++;
	}
      }
      else
      {
	itP2++;
      }
    }
    itP++;
  }

//    if(agent.getUniformNumber() == 1)
//    {
//      cout << "Post known--unknown-fix / Pre unkown-unknown-fx." << endl;
//      printList(players);
//    }

  // Time to remove excess players.
  while(players.size() > 10)
  {
    double minDist = 10000000000.0;
    itP = players.begin();
    std::list<PlayerState>::iterator mP;
    std::list<PlayerState>::iterator mP2;
    while(itP != players.end())
    {
      int unum1 = itP->getUniformNumber();
      std::list<PlayerState>::iterator itP2 = itP;
      itP2++;
      while(itP2 != players.end())
      {
	TVector delta = itP2->getPosition() - itP->getPosition();
	int unum2 = itP2->getUniformNumber();
	if( unum1 == -1 || unum2 == -1 )
	{
	  double curDist = delta*delta;
	  if(curDist < minDist)
	  {
	    minDist = curDist;
	    mP = itP;
	    mP2 = itP2;
	  }
	  else
	  {
	    itP2++;
	  }
	}
	else
	{
	  itP2++;
	}
      }
      itP++;
    }
    *mP = *mP2;
    players.erase(mP2);
  }

//    if(agent.getUniformNumber() == 1)
//    {
//      cout << "Post unkown-unknown-fx." << endl;
//      printList(players);
//    }

}

/**
 */
void WorldState::mergePlayers(std::vector<PlayerState>& mineV,
			      const std::vector<PlayerState>& otherV)
{
  // FIXME: Convert to linked list.

  std::list<PlayerState> result;
  std::list<PlayerState> mine;
  std::list<PlayerState>::iterator itMine;
  std::list<PlayerState> other;
  std::list<PlayerState>::iterator itOther;

  for(unsigned int i = 0; i < mineV.size(); i++)
    mine.push_back(mineV[i]);
  for(unsigned int i = 0; i < otherV.size(); i++)
    other.push_back(otherV[i]);

  fixConsitency(other);

//    if(otherV.size() > 10 && agent.getUniformNumber() == 1)
//    {
//      cout << "other vector tooo gawdamn large, " << otherV.size() << endl;
//      printList(other);
//    }

  // 1) Match by uniform number only.
  for(itOther = other.begin(); itOther != other.end();)
  {
    bool found = false;
    if(itOther->getUniformNumber() != -1)
    {
      for(itMine = mine.begin(); itMine != mine.end(); itMine++)
      {
	if(itMine->getUniformNumber() == itOther->getUniformNumber())
	{
	  found = true;
	  result.push_back(*itMine);
	  result.back().merge(*itOther);
	  itOther = other.erase(itOther);

	  itMine = mine.erase(itMine);

	  break;
	}
      }
    }
    if(!found)
      itOther++;
  }

  // 2) Match known uniform numbers to unkown ones.
  for(itOther = other.begin(); itOther != other.end();  )
  {
    double bestMatchVal = -1.0;
    std::list<PlayerState>::iterator bestMatch;
    int otherUnum = itOther->getUniformNumber();

    for(itMine = mine.begin(); itMine != mine.end(); itMine++)
    {
      int mineUnum = itMine->getUniformNumber();

      if( (mineUnum == -1 && otherUnum != -1)
	  || (mineUnum != -1 && otherUnum == -1) )
      {
	double curMatchVal = matchPlayer(*itMine, *itOther);
	if(curMatchVal > bestMatchVal)
	{
	  bestMatchVal = curMatchVal;
	  bestMatch = itMine;
	}
      }
    }

    if(bestMatchVal > 0.3)
    {
      result.push_back(*bestMatch);
      result.back().merge(*itOther);
      itOther = other.erase(itOther);
      mine.erase(bestMatch);
    }
    else
    {
      itOther++;
    }
  }

  // 3) Match unknown uniform numbers to unkown ones.
  for(itOther = other.begin(); itOther != other.end();  )
  {
    double bestMatchVal = -1.0;
    std::list<PlayerState>::iterator bestMatch;
    int otherUnum = itOther->getUniformNumber();

    for(itMine = mine.begin(); itMine != mine.end(); itMine++)
    {
      int mineUnum = itMine->getUniformNumber();

      if( mineUnum == -1 && otherUnum == -1 )
      {
	double curMatchVal = matchPlayer(*itMine, *itOther);
	if(curMatchVal > bestMatchVal)
	{
	  bestMatchVal = curMatchVal;
	  bestMatch = itMine;
	}
      }
    }

    if(bestMatchVal > 0.3)
    {
      result.push_back(*bestMatch);
      result.back().merge(*itOther);
      itOther = other.erase(itOther);
      mine.erase(bestMatch);
    }
    else
    {
      itOther++;
    }
  }

  // 4) Insert leftovers.
  if(other.size() > 0 && mine.size() > 0)
  {
//      printList(mine);
//      printList(other);
  }
  for(itOther = other.begin(); itOther != other.end(); itOther++)
    result.push_back(*itOther);
  for(itMine = mine.begin(); itMine != mine.end(); itMine++)
    result.push_back(*itMine);

//    if(agent.getUniformNumber() == 1)
//    {
//      cout << "Time(" << getGameTime() << ") : " << endl;
//      printList(result);
//      if(result.size() > 11)
//        cout << "Fuuuudge, shadows detected." << endl;
//    }

  // Check some stuff.
  bool found[11];
  for(int i = 0; i < 11; i++)
    found[i] = false;
  
  for(itMine = result.begin(); itMine != result.end(); itMine++) {
    int uNum = itMine->getUniformNumber();
    
    if(uNum != -1 && (uNum < 1 || uNum > 11))
      cout << "Wierd uNum found: " << uNum << endl;
    
    if(uNum != -1) {
      if(found[uNum-1]) 
	cout << "Dup. detected, how can this be???" << endl;
      found[uNum-1] = true;
    }
  }
  
  // Convert to vector.
  mineV.clear();
  for(itMine = result.begin(); itMine != result.end(); itMine++)
    mineV.push_back(*itMine);
}

/**
 */
std::vector<PlayerState>::iterator
WorldState::findMatch(PlayerState plr, std::vector<PlayerState>& matchList)
{
  double bestMatchVal = -1;
  std::vector<PlayerState>::iterator bestMatch = matchList.begin();
  std::vector<PlayerState>::iterator it;
  
  // Find match, if any.
  for(it = matchList.begin(); it != matchList.end(); it++)
  {
    double curMatchVal = matchPlayer(*it, plr);
    if(curMatchVal > bestMatchVal && curMatchVal > 0.3)
    {
      bestMatchVal = curMatchVal;
      bestMatch = it;
    }
  }
  return bestMatch;
}

/**
 */
void WorldState::clearCone(TVector point, float lDir, float rDir)
{

  if(point.getC() < 0.1)
    return;

  // Normalvectors to the two boundarylines.
  TVector lnVec = TVector::TPolarVector(1, lDir + 90);
  TVector rnVec = TVector::TPolarVector(1, rDir - 90);

  for(unsigned int i = 0; i < opponents.size(); i++)
  {
    TVector delta = opponents[i].getPosition() - point;
    if(delta * lnVec > 0 && delta * rnVec > 0
       && delta.getLength() < parameterView->SP_team_far_length)
    {
      // The opp is in the cone.
      //        opponents[i].setPosition(opponents[i].getPosition().setC(0.0));
      //        opponents[i].setSpeed(opponents[i].getSpeed().setC(0.0));
      opponents[i].setC(opponents[i].getC() * 0.6);
    }
  }

  for(unsigned int i = 0; i < teammates.size(); i++)
  {
    //    cout << i << ":" << teammates[i].getPosition() << endl;
    TVector delta = teammates[i].getPosition() - point;
    if(delta * lnVec > 0 && delta * rnVec > 0
       && delta.getLength() < parameterView->SP_team_far_length)
    {
      // The mate is in the cone.
      //        teammates[i].setPosition(teammates[i].getPosition().setC(0.0));
      //        teammates[i].setSpeed(teammates[i].getSpeed().setC(0.0));
      teammates[i].setC(teammates[i].getC() * 0.6);
    }
  }

  TVector delta = ball.getPosition() - point;
  if(delta * lnVec > 0 && delta * rnVec > 0)
  {
    // The opp is in the cone.
    //      ball.setPosition(ball.getPosition().setC(0.0));
    //      ball.setSpeed(ball.getSpeed().setC(0.0));
    ball.setC(0.0);
  }
}

// Merge two worldstates.
/**
 */
void WorldState::merge(const WorldState& other)
{
  // Synchronize the worldstates.
  WorldState curOther = other;
  while(curOther.getGameTime() < gameTime)
  {
    curOther.step();
  }

  while(curOther.getGameTime() > gameTime)
  {
    step();
  }

  curOther.reduceC(0.1);
  reduceC(0.1);

  std::vector<PlayerState>& otherTm = curOther.teammates;
  std::vector<PlayerState>& otherOp = curOther.opponents;

  // Find the agent in the other worldstate.
  if(matchPlayer(agent, curOther.getAgent()) > 0.99)
  {
    // Everything is fine and dandy, the agents are the same.
    if(agent.getC() < curOther.agent.getC())
    {
      // This shouldn't occur.. I think..
      agent = curOther.agent;
    }
  }
  else
  {
    std::vector<PlayerState>::iterator bestAgentMatch;
    // Match our agent to their teammates.
    bestAgentMatch = findMatch(agent, otherTm);
    if(bestAgentMatch != otherTm.end())
    {
      // FIXME:
      // Insert framing compensation here.
      //	  agent.merge(*bestAgentMatch);
      // This shouldn't be better then our own version.
      //	  cout << "Removing my agent from their team." << endl;
      otherTm.erase(bestAgentMatch);
    }

    // Match their agent to our teammates.
    //      otherTm.insert(otherTm.end(), curOther.agent);
    //        cout << "their agNum = " << curOther.agent.getUniformNumber() 
    //  	   << " mine = " << agent.getUniformNumber() << endl;
    //        cout << "Adding their agent to their team." << endl;
    otherTm.push_back(curOther.agent);
    //      curOther.print();
  }

  mergePlayers(opponents, otherOp);
  mergePlayers(teammates, otherTm);

  ball.setPosition(ball.getPosition().confSelect(curOther.ball.getPosition()));
//    cout << "Merge-pre : " << ball.getSpeed().getC() << endl;
  ball.setSpeed(ball.getSpeed().confSelect(curOther.ball.getSpeed()));
//    cout << "Merge-post: " << ball.getSpeed().getC() << endl;
}

// Merge two worldstates.
/**
 */
void WorldState::merge(TVector obsPoint, float lDir, float rDir, 
		       const WorldState& other)
{
  // Dobido.
  // And just add observations.

  // Do framing compensation.
  //  framingCompensate(other);

  //  cout << "New observations." << endl;
  clearCone(obsPoint, lDir, rDir);
  merge(other);

}

// Reduce the worldstate.
/**
 */
void WorldState::reduceC(float conf)
{
  std::vector<PlayerState>::iterator it = opponents.begin();
  while(it != opponents.end())
  {
    if(it->getPosition().getC() < conf)
    {
      it = opponents.erase(it);
    }
    else
    {
      it++;
    }
  }

  it = teammates.begin();
  while(it != teammates.end())
  {
    if(it->getPosition().getC() < conf)
    {
      it = teammates.erase(it);
    }
    else
    {
      it++;
    }
  }
}

/**
 */
void WorldState::reduceR(float radius)
{
  TVector agentPos = agent.getPosition();
  std::vector<PlayerState>::iterator it = opponents.begin();
  while(it != opponents.end())
  {
    TVector delta = it->getPosition() - agentPos;
    if(delta.getLength() > radius)
    {
      it = opponents.erase(it);
    }
    else
    {
      it++;
    }
  }

  it = teammates.begin();
  while(it != teammates.end())
  {
    TVector delta = it->getPosition() - agentPos;
    if(delta.getLength() > radius)
    {
      it = teammates.erase(it);
    }
    else
    {
      it++;
    }
  }
}

/**
 */
void WorldState::reduceCR(float conf, float radius)
{
  reduceC(conf);
  reduceR(radius);
}

/**
 */
bool WorldState::doKick(const TVector& v)
{
  return agent.doKick(v, ball);
}

/**
 */
bool WorldState::doDash(float power)
{
  return agent.doDash(power);
}

/**
 */
bool WorldState::doDashDistance(float dist)
{
  return agent.doDashDistance(dist);
}

/**
 */
bool WorldState::doTurn(float angle)
{
  return agent.doTurn(angle);
}

/**
 */
bool WorldState::doActualTurn(float angle)
{
  return agent.doActualTurn(angle);
}

/**
 */
bool WorldState::doTurnTo(float angle)
{
  return agent.doTurnTo(angle);
}

/**
 */
bool WorldState::doTurnTowards(const TVector& position)
{
  return agent.doTurnTowards(position);
}

/**
 */
bool WorldState::doTurnTowardsBall()
{
  BallState nextBall = ball;
  nextBall.step();
  return doTurnTowards(ball.getPosition());
}

/**
 */
bool WorldState::doTurnNeck(float angle)
{
  return agent.doTurnNeck(angle);
}

/**
 */
bool WorldState::doTurnNeckTo(float dir)
{
  return agent.doTurnNeckTo(dir);
}

/**
 */
bool WorldState::doTurnNeckTowards(const TVector& position)
{
  return agent.doTurnNeckTowards(position);
}

/**
 */
bool WorldState::doTurnNeckTowardsBall()
{
  BallState nextBall = ball;
  nextBall.step();
  return doTurnNeckTowards(ball.getPosition());
}

/**
 */
bool WorldState::doNull()
{
  return agent.doNull();
}


/**
 */
bool WorldState::doChangeViewWide()
{
  return agent.doChangeViewWide();
}

/**
 */
bool WorldState::doChangeViewNormal()
{
  return agent.doChangeViewNormal();
}

/**
 */
bool WorldState::doChangeViewNarrow()
{
  return agent.doChangeViewNarrow();
}

/**
 */
bool WorldState::doChangeViewHigh()
{
  return agent.doChangeViewHigh();
}

/**
 */
bool WorldState::doChangeViewLow()
{
  return agent.doChangeViewLow();
}


/**
 */
bool WorldState::doKickToSpeed(const TVector& speed)
{
  return agent.doKickToSpeed(speed, ball);
}

/**
 */
bool WorldState::doKickToPosition(const TVector& pos)
{
  return agent.doKickToPosition(pos, ball);
}

/**
 */
Step WorldState::getStep()
{
  return agent.getStep();
}

/**
 */
void WorldState::resolveCollisionHelp(PlayerState& plr1,
				      const PlayerState& plr2)
{
  // New hack. Just punish the colliding player by moving him back "speed"
  plr1.setPosition(plr1.getPosition()
		   - plr1.getSpeed() / parameterView->SP_player_decay);
  
  // User geometric solution as well if this isn't enough.
  TVector delta = plr2.getPosition() - plr1.getPosition();
  double overLap = 
    2.1 * parameterView->SP_player_size-
    delta.getLength();

  // If the distance between the object is to small
  // we just move them along X axis
  if(delta.getLength() < 0.001)
  {
    delta = TVector(1,0);
  }
  else
  {
    delta = delta.normalized();
  }

  delta = delta.normalized();
  plr1.setPosition(plr1.getPosition() - 1.05 * overLap * delta / 2);
  //  plr1.setSpeed(plr1.getSpeed() * -0.1);

}

/**
 */
void WorldState::resolveCollision(PlayerState& plr1,
				  PlayerState& plr2)
{
  PlayerState tmp = plr1;
  resolveCollisionHelp(plr1, plr2);
  resolveCollisionHelp(plr2, tmp);
}

/**
 */
void WorldState::resolveCollision(PlayerState& plr,
				  BallState& ball)
{
  //    // Copied from the sserver source
  //    //    Value r = size + obj.size ;
  //    double r = 1.05*(viewPack->parameterView->SP_player_size
  //  		   + viewPack->parameterView->SP_ball_size);
  
//    //    PVector dif = (pos - obj.pos) ;
//    TVector dif = plr.getPosition() - ball.getPosition();

//  //    Value d = pos.distance(obj.pos) ;
//    double d = dif.getLength();

//  //    Angle th = fabs(dif.angle(vel)) ;
//    double th = 
//      fabs(TVector(dif).rotate(-plr.getSpeed().getAngle()).getAngle());
//    //  fabs(TVector(plr1.getSpeed()).rotate(-dif.getAngle()).getAngle());
//    // I Hope... hehe...

//  //    Value l1 = d * cos(th) ;
//    TVector tmp = TVector::TPolarVector(d, th);
//    double l1 = tmp.getX();

//  //    Value h = d * sin(th) ;
//    double h = tmp.getY();

//  //    Value cosp = h / r ;
//    double cosp = h / r;

//  //    Value sinp = sqrt(1.0 - square(cosp)) ;
//    double sinp = sqrt(1.0 - cosp*cosp);

//  //    Value l2 = r * sinp ;
//    double l2 = r * sinp;

//  //    PVector dv = vel ;
//    TVector dv = plr.getSpeed();

//  //    dv.normalize(-(l1 + l2)) ;
//    dv = dv.normalized();

//  //    pos += dv ;

//    dv = -(l1+l2) * dv;
//    plr.setPosition(plr.getPosition() + dv);

  TVector delta = plr.getPosition() - ball.getPosition();
  double overLap = 
    1.05 * parameterView->SP_ball_size
    + 1.05 * parameterView->SP_player_size
    - delta.getLength();

  // If the distance between the object is to small
  // we just move them along X axis
  if(delta.getLength() < 0.001)
  {
    delta = TVector(1, 0);
  }
  else
  {
    delta = delta.normalized();
  }

  plr.setPosition(plr.getPosition() + 1.05 * overLap * delta / 2.0);
  plr.setSpeed(plr.getSpeed() * -0.1);
  ball.setPosition(ball.getPosition() - 1.05 * overLap * delta / 2.0);
  ball.setSpeed(plr.getSpeed() * -0.1);
}

/**
 */
void WorldState::checkCollisions(PlayerState& plr)
{
  bool collided = false;

  // Check against the agent.
  if(!(plr == agent))
  {
    if(agent.overlaps(plr))
    {
      collided = true;
      resolveCollision(plr, agent);
    }
  }
  
  if(agent.overlaps(ball))
  {
    // undo collision.
    collided = true;
    resolveCollision(plr, ball);
  }

  // Check against opponents.
  for(unsigned int i = 0; i < opponents.size(); i++)
  {
    if( !(opponents[i] == plr) )
    {
      if( opponents[i].overlaps(plr) )
      {
	// undo collision.
	//  	cout << "op:Plr coll" << endl;
	collided = true;
	resolveCollision(plr, opponents[i]);
      }
    }
  }

  // Check against teammates.
  for(unsigned int i = 0; i < teammates.size(); i++)
  {
    if( !(teammates[i] == plr) )
    {
      if( teammates[i].overlaps(plr) )
      {
	// undo collision.
	collided = true;
	resolveCollision(plr, teammates[i]);
      }
    }
  }

  if(collided)
  {
    plr.setSpeed(plr.getSpeed() * -0.1);
  }
}


/**
 */
void WorldState::checkCollisions(BallState& ball)
{
  bool collided = false;
  // Check against the agent.
  if(agent.overlaps(ball))
  {
    collided = true;
    resolveCollision(agent, ball);
  }

  // Check against opponents.
  for(unsigned int i = 0; i < opponents.size(); i++)
  {
    if( opponents[i].overlaps(ball) )
    {
      collided = true;
      resolveCollision(opponents[i], ball);
    }
  }
  
  // Check against teammates.
  for(unsigned int i = 0; i < teammates.size(); i++)
  {
    if( teammates[i].overlaps(ball) )
    {
      collided = true;
      resolveCollision(teammates[i], ball);
    }
  }
  
  if(collided)
  {
    ball.setSpeed(ball.getSpeed() * -0.1);
  }
}

/**
 */
void WorldState::checkCollisions()
{
  // First check agent.
  checkCollisions(agent);
  // Then ball.
  checkCollisions(ball);
}

/**
 */
const PlayerState& WorldState::getTheirInterceptor() const
{
  static PlayerState dummy;
  dummy.setPosition(TVector(-999, -999));
  PlayerState& interceptor = dummy;

  int bestTime = 1000;
  std::vector<PlayerState>::const_iterator plr;
  for(plr = opponents.begin(); plr != opponents.end(); plr++) {
    int curTime = plr->getInterceptTime(ball);
    if(curTime < bestTime) {
      interceptor = *plr;
      bestTime = curTime;
    }
  }
  return interceptor;
}

/**
 */
const PlayerState& WorldState::getOurInterceptor() const
{
  static PlayerState dummy;
  dummy.setPosition(TVector(-999, -999));
  PlayerState& interceptor = dummy;

  int bestTime = 1000;
  std::vector<PlayerState>::const_iterator plr;
  for(plr = teammates.begin(); plr != teammates.end(); plr++) {
    int curTime = plr->getInterceptTime(ball);
    if(curTime < bestTime) {
      interceptor = *plr;
      bestTime = curTime;
    }
  }
  if(agent.getInterceptTime(ball) < bestTime) {
    interceptor = agent;
  }

  return interceptor;
}

/** C++ er domt, you cannot nest functions
 * This one is private to getInterceptors
 */
bool reachesBall(const PlayerState& player,
		 const TVector& ballPosition,
		 const int ballStep,
		 const ParameterView* parameterView)
{
  TVector delta = ballPosition - player.getPosition();
  return delta.getLength() < (ballStep * parameterView->SP_player_speed_max
			      + parameterView->SP_kickable_area);
}

/** C++ er domt, you cannot nest functions
 * This one is private to getInterceptors
 */
int coarseInterceptionTime(const PlayerState& player,
			   const TVector& ballPosition,
			   const ParameterView* parameterView)
{
  TVector delta = ballPosition - player.getPosition();
  return (int)(1.1 * delta.getLength() / parameterView->SP_player_speed_max);
}


/**
 * Gets the first interceptors of each team. Will modify all
 * arguments. If there are no players of one kind, it's interception
 * time will be set to -1.
 */
void WorldState::getInterceptors(PlayerState& ourInterceptor,
				 int& ourInterceptionTime,
				 PlayerState& theirInterceptor,
				 int& theirInterceptionTime) const
{
  ourInterceptionTime = -1;
  theirInterceptionTime = -1;

  const ParameterView* parameterView = parameterView;

  std::vector<PlayerState>::const_iterator player;

  BallState ball = getBall();
  const int maxBallSteps = 20;
  int ballStep = 0;
  while(ballStep < maxBallSteps
	&& (ourInterceptionTime == -1
	    || theirInterceptionTime == -1))
  {
    if(ourInterceptionTime == -1)
    {
      if(reachesBall(agent, ball.getPosition(), ballStep, parameterView))
      {
	ourInterceptionTime = ballStep;
	ourInterceptor = agent;
      }
      for(player = teammates.begin(); player != teammates.end(); player++)
	if(reachesBall(*player, ball.getPosition(), ballStep, parameterView))
	{
	  ourInterceptionTime = ballStep;
	  ourInterceptor = *player;
	}
    }

    if(theirInterceptionTime == -1)
    {
      for(player = opponents.begin(); player != opponents.end(); player++)
	if(reachesBall(*player, ball.getPosition(), ballStep, parameterView))
	{
	  theirInterceptionTime = ballStep;
	  theirInterceptor = *player;
	}
    }
    ball.step();
    ballStep++;
  }

  // If no interceptors were found, use a coarse heuristic

  if(ourInterceptionTime == -1)
  {
    ourInterceptionTime = coarseInterceptionTime(agent, ball.getPosition(),
						 parameterView);
    ourInterceptor = agent;

    for(player = teammates.begin(); player != teammates.end(); player++)
    {
      int time = coarseInterceptionTime(*player, ball.getPosition(),
					parameterView);
      if(time < ourInterceptionTime)
      {
	ourInterceptionTime = time;
	ourInterceptor = *player;
      }
    }
  }

  if(theirInterceptionTime == -1)
  {
    for(player = opponents.begin(); player != opponents.end(); player++)
    {
      int time = coarseInterceptionTime(*player, ball.getPosition(),
					parameterView);
      if(time < theirInterceptionTime
	 || theirInterceptionTime == -1)
      {
	theirInterceptionTime = time;
	theirInterceptor = *player;
      }
    }
  }
}

/** Models a player, the agent in the given world state, and returns
 * it.
 */
PlayerState WorldState::modelPlayer(WorldState& world) {
  // NOT IMPLEMENTED!
  return world.getAgent();
}

/** Iterates through all teammates and opponents and models
 * their actions. 
 */
void WorldState::model() {
//    WorldState modelWorld(*this);
  
//    std::vector<PlayerState>::iterator plr;
//    for(plr = teammates.begin(); plr != teammates.end(); plr++) {
//      modelWorld.changeAgent(*plr);
//      *plr = modelPlayer(modelWorld);
//    }
//    //   for(plr = opponents.begin(); plr != opponents.end(); plr++) {
//    //     modelWorld.changeAgent(*plr);
//    //     *plr = modelPlayer(modelWorld);
//    //   }
}

// From State
/**
 */
void WorldState::step()
{
  gameTime++;
//    model();

  // Update opponents
  for(unsigned int i = 0; i < opponents.size(); i++) 
    while(opponents[i].getGameTime() < gameTime) 
      opponents[i].step();

  // Update opponents
  for(unsigned int i = 0; i < teammates.size(); i++)
    while(teammates[i].getGameTime() < gameTime)
      teammates[i].step();

  // Update the ball
  while(ball.getGameTime() < gameTime)
    ball.step();

  // Update the agent
  while(agent.getGameTime() < gameTime)
    agent.step();

  // Check collisions.
  checkCollisions();
}

/** Radial step from the agent. stepDistance is the distance for
 * one step. Inside stepDistance, the step is 1, inside
 * 2 x stepDistance the step is 2, and so on.
 * The agent is not stepped.
 */
void WorldState::stepR(float stepDistance) {
  stepR(stepDistance, agent.getPosition());
}

/** Radial step from a given center position. stepDistance is the
 * distance for one step. Inside stepDistance, the step is 1, inside
 * 2 x stepDistance the step is 2, and so on.
 * The agent is not stepped.
 */
void WorldState::stepR(float stepDistance, const TVector& center) {
  
  // initial step
  step();
  
  // build distance vectors
  std::vector<float> tmDist(teammates.size());
  std::vector<float> opDist(opponents.size());
  float ballDist;
  float maxDist(0);
  
  for(unsigned int i = 0; i < teammates.size(); i++) {
    tmDist[i] = (teammates[i].getPosition() - center).getLength();
    if(tmDist[i] > maxDist)
      maxDist = tmDist[i];
  }

  for(unsigned int i = 0; i < opponents.size(); i++) {
    opDist[i] = (opponents[i].getPosition() - center).getLength();
    if(opDist[i] > maxDist)
      maxDist = opDist[i];
  }
 
  ballDist = (ball.getPosition() - center).getLength();

  WorldState modelWorld;
  for(float r = stepDistance; r <= maxDist; r += stepDistance) {
    // ** Model **
    modelWorld = *this;
    
    // teammates
    for(unsigned int i = 0; i < teammates.size(); i++) {
      if(tmDist[i] > r) {
	modelWorld.changeAgent(teammates[i]);
	teammates[i] = modelPlayer(modelWorld);
      }
    }

    // opponents
    for(unsigned int i = 0; i < opponents.size(); i++) {
      if(tmDist[i] > r) {
	modelWorld.changeAgent(opponents[i]);
	opponents[i] = modelPlayer(modelWorld);
      }
    }
    
    // ** Step **

    // teammates
    for(unsigned int i = 0; i < teammates.size(); i++) {
      if(tmDist[i] > r) 
	teammates[i].step();
    }

    // opponents
    for(unsigned int i = 0; i < opponents.size(); i++) {
      if(tmDist[i] > r) 
	opponents[i].step();
    }

    // ball
    if(ballDist > r)
      ball.step();
  }
}

/**
 */
void printPlayer(const PlayerState& plr)
{
  //    cout << "Player: ==========" << endl
  //         << "Pos: " << plr.getPosition() << endl
  //         << "Spd: " << plr.getSpeed() << endl
  //         << "Dir: " << plr.getDirection() << endl
  //         << "Unm: " << plr.getUniformNumber() << endl;
  cout << plr.getUniformNumber() << ", ";
}

/**
 */
void WorldState::print()
{      
  //    cout << "==================" << endl
  //         << "Time: " << getGameTime() << endl
  //         << "Teammates: " << getTeammates().size() << endl
  //         << "Opponents: " << getOpponents().size() << endl
  //         << "Ball: ============" << endl
  //         << "Pos:" << getBall().getPosition() << endl
  //         << "Spd:" << getBall().getSpeed() << endl
  //         << "Acc:" << getBall().getAcceleration() << endl
  //         << "Agent:  ==========" << endl
  //         << "Stm: " << getAgent().getStamina() << " "
  //         << "Rec: " << getAgent().getRecovery() << " "
  //         << "Eff: " << getAgent().getEffort() << " " << endl;
  //    printPlayer(getAgent());

  cout << "Teammates:  ======" << endl << "{";
  for(unsigned int i = 0; i < teammates.size(); i++)
  {
    printPlayer(teammates[i]);
  }
  cout << "}" << endl;

  cout << "Opponents:  ======" << endl << "{";
  for(unsigned int i = 0; i < opponents.size(); i++)
  {
    printPlayer(opponents[i]);
  }
  cout << "}" << endl;
}

// Returns number of chars needed.
/**
 */
std::string WorldState::serialize() const
{
  std::string result; short Short; char Char;

  // Gametime and number of teammates/opponents.
  result += asString(Short = gameTime);
  result += asString(Char = teammates.size());
  result += asString(Char = opponents.size());

  // Ball
  result += ball.serialize();

  // Agent
  result += agent.serialize();

  // Teammates
  std::vector<PlayerState>::const_iterator plr;
  for(plr = teammates.begin(); plr != teammates.end(); plr++)
    result += plr->serialize();

  // Opponents
  for(plr = opponents.begin(); plr != opponents.end(); plr++)
    result += plr->serialize();

  return result;
}

/** Transforms the worldstate to one where the argument is
 * the new agent. The argument should be a playerstate that is
 * identical to one in the WorldState either the old agent, one
 * of the teammates or one of the opponents. The worldstate will
 * seem as the one for that player. Teams will be changed and the 
 * field will be mirrored etc.
 */
void WorldState::changeAgent(const PlayerState& newAgent)
{
  if(newAgent == agent) 
    return;
  
  // Search through teammates.
  std::vector<PlayerState>::iterator it;
  for(it = teammates.begin(); it != teammates.end(); it++)
  {
    if(*it == newAgent)
    {
      // We found the player in the teammates list.
      *it = agent;
      agent = newAgent;
      return;
    }
  }

  // The new agent isn't on our team, so we switch sides.
  swapSides(); // Switching team

  if(newAgent == agent)
    return;

  // Now the former opponents are our teammates.
  for(it = teammates.begin(); it != teammates.end(); it++)
  {
    if(*it == newAgent)
    {
      // We found the player in the teammates list.
      *it = agent;
      agent = newAgent;
      return;
    }
  }

  cout << "WARNING! tried to changeAgent with a player"
       << " that isn't in the worldstate." << endl;
  swapSides(); // Restore everything
}

void WorldState::mirror(PlayerState& plr)
{
  plr.setSpeed(-plr.getSpeed());
  plr.setPosition(-plr.getPosition());
  plr.setDirection(180 + plr.getDirection());
}

/** Changes the teams in the worldstate. The last member of either
 * team will become the agent.
 */
void WorldState::swapSides()
{
  teammates.push_back(agent);
  teammates.swap(opponents);
  agent = teammates.back();
  teammates.pop_back();

  // Swap team
  team = (team == T_OUR_TEAM ? T_THEIR_TEAM : T_OUR_TEAM);

  // Mirror the entire field.
  ball.setSpeed(-ball.getSpeed());
  ball.setPosition(-ball.getPosition());
  mirror(agent);

  std::vector<PlayerState>::iterator it;
  for(it = teammates.begin(); it != teammates.end(); it++)
  {
    mirror(*it);
  }

  for(it = opponents.begin(); it != opponents.end(); it++)
  {
    mirror(*it);
  }
}

// Returns number of chars consumed.
/**
 */
std::string WorldState::deserialize(const std::string& buf)
{
  std::string result = buf; short Short(0); char Char(0);

  int numOps, numTms;

  // Gametime and number of teammates/opponents.
  gameTime = fromString(Short, result);
  result = skipString(Short, result);
  numTms = fromString(Char, result);
  result = skipString(Char, result);
  numOps = fromString(Char, result);
  result = skipString(Char, result);

  // Ball
  result = ball.deserialize(result);

  // Agent
  result = agent.deserialize(result);

  // Teammates
  teammates.clear();
  for(int i = 0; i < numTms; i++)
  {
    PlayerState plr;
    result = plr.deserialize(result);
    teammates.push_back(plr);
  }

  // Opponents
  opponents.clear();
  for(int i = 0; i < numOps; i++)
  {
    PlayerState plr;
    result = plr.deserialize(result);
    opponents.push_back(plr);
  }
  
  return result;
}

float WorldState::getOffsideLine() const
{
  // Stolen from PassTree
  // FIXME: Is the penalty-area stuff correct?
  const float theirPenaltyArea = 52.5 - 16.5;

  float ourOffsideLine = 0; 
  float maxX = 0;
  float x;
  const std::vector<PlayerState>& op = getOpponents();
  std::vector<PlayerState>::const_iterator plr;
  for(plr = op.begin(); plr != op.end(); plr++) {
    x = plr->getPosition().getX();
    if(x > ourOffsideLine) 
      if(x > maxX) {
	ourOffsideLine = maxX;
	maxX = x;
      } else 
	ourOffsideLine = x;
  }
  if(ourOffsideLine > theirPenaltyArea)
    ourOffsideLine = theirPenaltyArea;

  return ourOffsideLine;
}

bool WorldState::isPlayerOffside(const PlayerState& player) const
{
  return player.getPosition().getX() < getOffsideLine();
}

RS_END_NAMESPACE

